# Copyright (c) 2015  aggftw@gmail.com
# Distributed under the terms of the Modified BSD License.

from mock import patch, PropertyMock, MagicMock
import pytest
import requests
from requests_kerberos.kerberos_ import HTTPKerberosAuth, REQUIRED, OPTIONAL
from sparkmagic.auth.basic import Basic
from sparkmagic.auth.kerberos import Kerberos
from sparkmagic.auth.customauth import Authenticator
from sparkmagic.livyclientlib.endpoint import Endpoint
from sparkmagic.livyclientlib.exceptions import (
    HttpClientException,
    HttpSessionAdapterConfigException,
)
from sparkmagic.livyclientlib.exceptions import BadUserConfigurationException
from sparkmagic.livyclientlib.linearretrypolicy import LinearRetryPolicy
from sparkmagic.livyclientlib.reliablehttpclient import ReliableHttpClient
import sparkmagic.utils.configuration as conf
import sparkmagic.utils.constants as constants

retry_policy = None
sequential_values = []
basic_auth = Basic()
kerberos_auth = Kerberos()
endpoint = Endpoint("http://url.com", basic_auth)


def setup_function():
    global retry_policy
    retry_policy = LinearRetryPolicy(0.01, 5)


def teardown_function():
    pass


def return_sequential():
    global sequential_values
    val = sequential_values[0]
    sequential_values = sequential_values[1:]
    return val


def test_compose_url():
    client = ReliableHttpClient(endpoint, {}, retry_policy)

    composed = client.compose_url("r")
    assert "http://url.com/r" == composed

    composed = client.compose_url("/r")
    assert "http://url.com/r" == composed

    client = ReliableHttpClient(endpoint, {}, retry_policy)

    composed = client.compose_url("r")
    assert "http://url.com/r" == composed

    composed = client.compose_url("/r")
    assert "http://url.com/r" == composed


def test_get():
    with patch("requests.Session.get") as patched_get:
        type(patched_get.return_value).status_code = 200

        client = ReliableHttpClient(endpoint, {}, retry_policy)

        result = client.get("r", [200])

        assert 200 == result.status_code


def test_get_throws():
    with pytest.raises(HttpClientException):
        with patch("requests.Session.get") as patched_get:
            type(patched_get.return_value).status_code = 500

            client = ReliableHttpClient(endpoint, {}, retry_policy)

            client.get("r", [200])


def test_get_will_retry():
    global sequential_values, retry_policy
    retry_policy = MagicMock()
    retry_policy.should_retry.return_value = True
    retry_policy.seconds_to_sleep.return_value = 0.01

    with patch("requests.Session.get") as patched_get:
        # When we call assert_equals in this unit test, the side_effect is executed.
        # So, the last status_code should be repeated.
        sequential_values = [500, 200, 200]
        pm = PropertyMock()
        pm.side_effect = return_sequential
        type(patched_get.return_value).status_code = pm
        client = ReliableHttpClient(endpoint, {}, retry_policy)

        result = client.get("r", [200])

        assert 200 == result.status_code
        retry_policy.should_retry.assert_called_once_with(500, False, 0)
        retry_policy.seconds_to_sleep.assert_called_once_with(0)


def test_post():
    with patch("requests.Session.post") as patched_post:
        type(patched_post.return_value).status_code = 200

        client = ReliableHttpClient(endpoint, {}, retry_policy)

        result = client.post("r", [200], {})

        assert 200 == result.status_code


def test_post_throws():
    with pytest.raises(HttpClientException):
        with patch("requests.Session.post") as patched_post:
            type(patched_post.return_value).status_code = 500

            client = ReliableHttpClient(endpoint, {}, retry_policy)

            client.post("r", [200], {})


def test_post_will_retry():
    global sequential_values, retry_policy
    retry_policy = MagicMock()
    retry_policy.should_retry.return_value = True
    retry_policy.seconds_to_sleep.return_value = 0.01

    with patch("requests.Session.post") as patched_post:
        # When we call assert_equals in this unit test, the side_effect is executed.
        # So, the last status_code should be repeated.
        sequential_values = [500, 200, 200]
        pm = PropertyMock()
        pm.side_effect = return_sequential
        type(patched_post.return_value).status_code = pm
        client = ReliableHttpClient(endpoint, {}, retry_policy)

        result = client.post("r", [200], {})

        assert 200 == result.status_code
        retry_policy.should_retry.assert_called_once_with(500, False, 0)
        retry_policy.seconds_to_sleep.assert_called_once_with(0)


def test_delete():
    with patch("requests.Session.delete") as patched_delete:
        type(patched_delete.return_value).status_code = 200

        client = ReliableHttpClient(endpoint, {}, retry_policy)

        result = client.delete("r", [200])

        assert 200 == result.status_code


def test_delete_throws():
    with pytest.raises(HttpClientException):
        with patch("requests.Session.delete") as patched_delete:
            type(patched_delete.return_value).status_code = 500

            client = ReliableHttpClient(endpoint, {}, retry_policy)

            client.delete("r", [200])


def test_delete_will_retry():
    global sequential_values, retry_policy
    retry_policy = MagicMock()
    retry_policy.should_retry.return_value = True
    retry_policy.seconds_to_sleep.return_value = 0.01

    with patch("requests.Session.delete") as patched_delete:
        # When we call assert_equals in this unit test, the side_effect is executed.
        # So, the last status_code should be repeated.
        sequential_values = [500, 200, 200]
        pm = PropertyMock()
        pm.side_effect = return_sequential
        type(patched_delete.return_value).status_code = pm
        client = ReliableHttpClient(endpoint, {}, retry_policy)

        result = client.delete("r", [200])

        assert 200 == result.status_code
        retry_policy.should_retry.assert_called_once_with(500, False, 0)
        retry_policy.seconds_to_sleep.assert_called_once_with(0)


def test_will_retry_error_no():
    global sequential_values, retry_policy
    retry_policy = MagicMock()
    retry_policy.should_retry.return_value = False
    retry_policy.seconds_to_sleep.return_value = 0.01

    with patch("requests.Session.get") as patched_get:
        patched_get.side_effect = requests.exceptions.ConnectionError()
        client = ReliableHttpClient(endpoint, {}, retry_policy)

        try:
            client.get("r", [200])
            assert False
        except HttpClientException:
            retry_policy.should_retry.assert_called_once_with(None, True, 0)


def test_basic_auth_check_auth():
    endpoint = Endpoint("http://url.com", basic_auth)
    client = ReliableHttpClient(endpoint, {}, retry_policy)
    assert isinstance(client._auth, Basic)
    assert hasattr(client._auth, "username")
    assert hasattr(client._auth, "password")
    assert client._auth.username == endpoint.auth.username
    assert client._auth.password == endpoint.auth.password


def test_no_auth_check_auth():
    endpoint = Endpoint("http://url.com", None)
    client = ReliableHttpClient(endpoint, {}, retry_policy)
    assert client._auth == None


def test_kerberos_auth_check_auth():
    endpoint = Endpoint("http://url.com", kerberos_auth)
    client = ReliableHttpClient(endpoint, {}, retry_policy)
    assert client._auth is not None
    assert isinstance(client._auth, HTTPKerberosAuth)
    assert hasattr(client._auth, "mutual_authentication")
    assert client._auth.mutual_authentication == REQUIRED


def test_kerberos_auth_custom_configuration():
    custom_kerberos_conf = {"mutual_authentication": OPTIONAL, "force_preemptive": True}
    overrides = {conf.kerberos_auth_configuration.__name__: custom_kerberos_conf}
    conf.override_all(overrides)
    kerberos_auth = Kerberos()
    endpoint = Endpoint("http://url.com", kerberos_auth)
    client = ReliableHttpClient(endpoint, {}, retry_policy)
    assert client._auth is not None
    assert isinstance(client._auth, HTTPKerberosAuth)
    assert hasattr(client._auth, "mutual_authentication")
    assert client._auth.mutual_authentication == OPTIONAL
    assert hasattr(client._auth, "force_preemptive")
    assert client._auth.force_preemptive == True


def test_http_custom_session_config():
    http_abc_com = "http://abc.com"
    https_xyz_com = "https://xyz.com"
    custom_session_conf = {
        "adapters": [
            {"prefix": http_abc_com, "adapter": "requests.adapters.BaseAdapter"},
            {"prefix": https_xyz_com, "adapter": "requests.adapters.HTTPAdapter"},
        ]
    }
    overrides = {conf.http_session_config.__name__: custom_session_conf}
    conf.override_all(overrides)
    kerberos_auth = Kerberos()
    endpoint = Endpoint("http://url.com", kerberos_auth)
    client = ReliableHttpClient(endpoint, {}, retry_policy)
    assert client._session is not None
    assert client._session.adapters[http_abc_com] is not None
    assert isinstance(
        client._session.adapters[http_abc_com], requests.adapters.BaseAdapter
    )
    assert client._session.adapters[https_xyz_com] is not None
    assert isinstance(
        client._session.adapters[https_xyz_com], requests.adapters.HTTPAdapter
    )


def test_invalid_http_custom_session_config():
    http_abc_com = "http://abc.com"
    custom_session_conf = {"adapters": [{"prefix": http_abc_com}]}
    overrides = {conf.http_session_config.__name__: custom_session_conf}
    conf.override_all(overrides)
    kerberos_auth = Kerberos()
    endpoint = Endpoint("http://url.com", kerberos_auth)
    with pytest.raises(HttpSessionAdapterConfigException):
        ReliableHttpClient(endpoint, {}, retry_policy)



from mock import MagicMock
import pytest
from IPython.core.magic import magics_class

import sparkmagic.utils.configuration as conf
from sparkmagic.utils.utils import parse_argstring_or_throw, initialize_auth
import sparkmagic.utils.constants as constants
from sparkmagic.kernels.kernelmagics import KernelMagics, Namespace
from sparkmagic.magics.remotesparkmagics import RemoteSparkMagics
from sparkmagic.livyclientlib.exceptions import (
    LivyClientTimeoutException,
    BadUserDataException,
    LivyUnexpectedStatusException,
    SessionManagementException,
    HttpClientException,
    DataFrameParseException,
    SqlContextNotFoundException,
    SparkStatementException,
)
from sparkmagic.livyclientlib.endpoint import Endpoint
from sparkmagic.livyclientlib.command import Command
from sparkmagic.auth.basic import Basic
import importlib

magic = None
spark_controller = None
shell = None
ipython_display = MagicMock()
spark_events = None


@magics_class
class MockKernelMagics(KernelMagics):
    def __init__(self, shell, data=None, spark_events=None):
        super(MockKernelMagics, self).__init__(shell, spark_events=spark_events)

        self.language = constants.LANG_PYTHON
        self.endpoint = Endpoint("url", None)

    def refresh_configuration(self):
        self.endpoint = Endpoint("new_url", None)


def setup_function():
    global magic, spark_controller, shell, ipython_display, spark_events, conf

    conf.override_all({})
    spark_events = MagicMock()

    magic = MockKernelMagics(shell=None, spark_events=spark_events)
    magic.shell = shell = MagicMock()
    magic.ipython_display = ipython_display = MagicMock()
    magic.spark_controller = spark_controller = MagicMock()
    magic._generate_uuid = MagicMock(return_value="0000")


def teardown_function():
    pass


def test_local():
    with pytest.raises(NotImplementedError):
        magic.local("")


def test_start_session():
    line = ""
    assert not magic.session_started

    ret = magic._do_not_call_start_session(line)

    assert ret
    assert magic.session_started
    spark_controller.add_session.assert_called_once_with(
        magic.session_name,
        magic.endpoint,
        False,
        {"kind": constants.SESSION_KIND_PYSPARK},
    )

    # Call a second time
    ret = magic._do_not_call_start_session(line)
    assert ret
    assert magic.session_started
    assert spark_controller.add_session.call_count == 1


def test_start_session_times_out():
    line = ""
    spark_controller.add_session = MagicMock(side_effect=LivyClientTimeoutException)
    assert not magic.session_started

    ret = magic._do_not_call_start_session(line)

    assert not ret
    assert not magic.session_started
    assert magic.fatal_error
    assert ipython_display.send_error.call_count == 1

    # Call after fatal error
    ipython_display.send_error.reset_mock()
    ret = magic._do_not_call_start_session(line)
    assert not ret
    assert not magic.session_started
    assert ipython_display.send_error.call_count == 1


def test_start_session_times_out_all_errors_are_fatal():
    with pytest.raises(LivyClientTimeoutException):
        conf.override_all({"all_errors_are_fatal": True})

        line = ""
        spark_controller.add_session = MagicMock(side_effect=LivyClientTimeoutException)
        assert not magic.session_started

        _ = magic._do_not_call_start_session(line)


def test_delete_session():
    line = ""
    magic.session_started = True

    magic._do_not_call_delete_session(line)

    assert not magic.session_started
    spark_controller.delete_session_by_name.assert_called_once_with(magic.session_name)

    # Call a second time
    magic._do_not_call_delete_session(line)
    assert not magic.session_started
    assert spark_controller.delete_session_by_name.call_count == 1


def test_delete_session_expected_exception():
    line = ""
    magic.session_started = True
    spark_controller.delete_session_by_name.side_effect = BadUserDataException("hey")

    magic._do_not_call_delete_session(line)

    assert not magic.session_started
    spark_controller.delete_session_by_name.assert_called_once_with(magic.session_name)
    ipython_display.send_error.assert_called_once_with(
        constants.EXPECTED_ERROR_MSG.format(
            spark_controller.delete_session_by_name.side_effect
        )
    )


def test_change_language():
    language = constants.LANG_SCALA.upper()
    line = "-l {}".format(language)

    magic._do_not_call_change_language(line)

    assert constants.LANG_SCALA == magic.language
    assert Endpoint("new_url", None) == magic.endpoint


def test_change_language_session_started():
    language = constants.LANG_PYTHON
    line = "-l {}".format(language)
    magic.session_started = True

    magic._do_not_call_change_language(line)

    assert ipython_display.send_error.call_count == 1
    assert constants.LANG_PYTHON == magic.language
    assert Endpoint("url", None) == magic.endpoint


def test_change_language_not_valid():
    language = "not_valid"
    line = "-l {}".format(language)

    magic._do_not_call_change_language(line)

    assert ipython_display.send_error.call_count == 1
    assert constants.LANG_PYTHON == magic.language
    assert Endpoint("url", None) == magic.endpoint


def test_change_endpoint():
    s = "server"
    u = "user"
    p = "password"
    t = constants.AUTH_BASIC
    line = "-s {} -u {} -p {} -t {}".format(s, u, p, t)
    magic._do_not_call_change_endpoint(line)
    args = Namespace(
        auth="Basic_Access", password="password", url="server", user="user"
    )
    auth_instance = initialize_auth(args)
    endpoint = Endpoint(s, auth_instance)
    assert endpoint.url == magic.endpoint.url
    assert Endpoint(s, auth_instance) == magic.endpoint


def test_change_endpoint_session_started():
    with pytest.raises(BadUserDataException):
        u = "user"
        p = "password"
        s = "server"
        line = "-s {} -u {} -p {}".format(s, u, p)
        magic.session_started = True
        magic._do_not_call_change_endpoint(line)


def test_info():
    magic._print_endpoint_info = print_info_mock = MagicMock()
    line = ""
    session_info = [MagicMock(), MagicMock()]
    spark_controller.get_all_sessions_endpoint = MagicMock(return_value=session_info)
    magic.session_started = True

    magic.info(line)

    print_info_mock.assert_called_once_with(
        session_info, spark_controller.get_session_id_for_client.return_value
    )
    spark_controller.get_session_id_for_client.assert_called_once_with(
        magic.session_name
    )

    _assert_magic_successful_event_emitted_once("info")


def test_info_without_active_session():
    magic._print_endpoint_info = print_info_mock = MagicMock()
    line = ""
    session_info = [MagicMock(), MagicMock()]
    spark_controller.get_all_sessions_endpoint = MagicMock(return_value=session_info)

    magic.info(line)

    print_info_mock.assert_called_once_with(session_info, None)

    _assert_magic_successful_event_emitted_once("info")


def test_info_with_cell_content():
    magic._print_endpoint_info = print_info_mock = MagicMock()
    line = ""
    session_info = ["1", "2"]
    spark_controller.get_all_sessions_endpoint_info = MagicMock(
        return_value=session_info
    )
    error_msg = "Cell body for %%info magic must be empty; got 'howdy' instead"

    magic.info(line, cell="howdy")

    print_info_mock.assert_not_called()
    assert ipython_display.send_error.call_count == 1
    spark_controller.get_session_id_for_client.assert_not_called()
    _assert_magic_failure_event_emitted_once("info", BadUserDataException(error_msg))


def test_info_with_argument():
    magic._print_endpoint_info = print_info_mock = MagicMock()
    line = "hey"
    session_info = ["1", "2"]
    spark_controller.get_all_sessions_endpoint_info = MagicMock(
        return_value=session_info
    )

    magic.info(line)

    print_info_mock.assert_not_called()
    assert ipython_display.send_error.call_count == 1
    spark_controller.get_session_id_for_client.assert_not_called()


def test_info_unexpected_exception():
    magic._print_endpoint_info = MagicMock()
    line = ""
    spark_controller.get_all_sessions_endpoint = MagicMock(
        side_effect=ValueError("utter failure")
    )

    magic.info(line)
    _assert_magic_failure_event_emitted_once(
        "info", spark_controller.get_all_sessions_endpoint.side_effect
    )
    ipython_display.send_error.assert_called_once_with(
        constants.INTERNAL_ERROR_MSG.format(
            spark_controller.get_all_sessions_endpoint.side_effect
        )
    )


def test_info_expected_exception():
    magic._print_endpoint_info = MagicMock()
    line = ""
    spark_controller.get_all_sessions_endpoint = MagicMock(
        side_effect=SqlContextNotFoundException("utter failure")
    )

    magic.info(line)
    _assert_magic_failure_event_emitted_once(
        "info", spark_controller.get_all_sessions_endpoint.side_effect
    )
    ipython_display.send_error.assert_called_once_with(
        constants.EXPECTED_ERROR_MSG.format(
            spark_controller.get_all_sessions_endpoint.side_effect
        )
    )


def test_help():
    magic.help("")

    assert ipython_display.html.call_count == 1
    _assert_magic_successful_event_emitted_once("help")


def test_help_with_cell_content():
    msg = "Cell body for %%help magic must be empty; got 'HAHAH' instead"
    magic.help("", cell="HAHAH")

    assert ipython_display.send_error.call_count == 1
    assert ipython_display.html.call_count == 0
    _assert_magic_failure_event_emitted_once("help", BadUserDataException(msg))


def test_help_with_argument():
    magic.help("argument here")

    assert ipython_display.send_error.call_count == 1
    assert ipython_display.html.call_count == 0


def test_logs():
    logs = "logs"
    line = ""

    magic.logs(line)
    ipython_display.write.assert_called_once_with("No logs yet.")
    _assert_magic_successful_event_emitted_once("logs")

    ipython_display.write.reset_mock()

    magic.session_started = True

    spark_controller.get_logs = MagicMock(return_value=logs)
    magic.logs(line)
    ipython_display.write.assert_called_once_with(logs)
    spark_controller.get_logs.assert_called_once_with()


def test_logs_with_cell_content():
    logs = "logs"
    line = ""
    msg = "Cell body for %%logs magic must be empty; got 'BOOP' instead"

    magic.logs(line, cell="BOOP")
    assert ipython_display.send_error.call_count == 1
    _assert_magic_failure_event_emitted_once("logs", BadUserDataException(msg))


def test_logs_with_argument():
    line = "-h"

    magic.logs(line)
    assert ipython_display.send_error.call_count == 1


def test_logs_unexpected_exception():
    line = ""

    magic.session_started = True

    spark_controller.get_logs = MagicMock(
        side_effect=SyntaxError("There was some sort of error")
    )
    magic.logs(line)
    spark_controller.get_logs.assert_called_once_with()
    _assert_magic_failure_event_emitted_once(
        "logs", spark_controller.get_logs.side_effect
    )
    ipython_display.send_error.assert_called_once_with(
        constants.INTERNAL_ERROR_MSG.format(spark_controller.get_logs.side_effect)
    )


def test_logs_expected_exception():
    line = ""

    magic.session_started = True

    spark_controller.get_logs = MagicMock(
        side_effect=LivyUnexpectedStatusException("There was some sort of error")
    )
    magic.logs(line)
    spark_controller.get_logs.assert_called_once_with()
    _assert_magic_failure_event_emitted_once(
        "logs", spark_controller.get_logs.side_effect
    )
    ipython_display.send_error.assert_called_once_with(
        constants.EXPECTED_ERROR_MSG.format(spark_controller.get_logs.side_effect)
    )


def test_configure():
    # Mock info method
    magic.info = MagicMock()

    # Session not started
    conf.override_all({})
    magic.configure("", '{"extra": "yes"}')
    assert conf.session_configs() == {"extra": "yes"}
    _assert_magic_successful_event_emitted_once("configure")
    magic.info.assert_called_once_with("")

    # Session started - no -f
    magic.session_started = True
    conf.override_all({})
    magic.configure("", '{"extra": "yes"}')
    assert conf.session_configs() == {}
    assert ipython_display.send_error.call_count == 1

    # Session started - with -f
    magic.info.reset_mock()
    conf.override_all({})
    magic.configure("-f", '{"extra": "yes"}')
    assert conf.session_configs() == {"extra": "yes"}
    spark_controller.delete_session_by_name.assert_called_once_with(magic.session_name)
    spark_controller.add_session.assert_called_once_with(
        magic.session_name,
        magic.endpoint,
        False,
        {"kind": constants.SESSION_KIND_PYSPARK, "extra": "yes"},
    )
    magic.info.assert_called_once_with("")


def test_configure_unexpected_exception():
    magic.info = MagicMock()

    magic._override_session_settings = MagicMock(side_effect=ValueError("help"))
    magic.configure("", '{"extra": "yes"}')
    _assert_magic_failure_event_emitted_once(
        "configure", magic._override_session_settings.side_effect
    )
    ipython_display.send_error.assert_called_once_with(
        constants.INTERNAL_ERROR_MSG.format(
            magic._override_session_settings.side_effect
        )
    )


def test_configure_expected_exception():
    magic.info = MagicMock()

    magic._override_session_settings = MagicMock(
        side_effect=BadUserDataException("help")
    )
    magic.configure("", '{"extra": "yes"}')
    _assert_magic_failure_event_emitted_once(
        "configure", magic._override_session_settings.side_effect
    )
    ipython_display.send_error.assert_called_once_with(
        constants.EXPECTED_ERROR_MSG.format(
            magic._override_session_settings.side_effect
        )
    )


def test_configure_cant_parse_object_as_json():
    magic.info = MagicMock()

    magic._override_session_settings = MagicMock(
        side_effect=BadUserDataException("help")
    )
    magic.configure("", "I CAN'T PARSE THIS AS JSON")
    _assert_magic_successful_event_emitted_once("configure")
    assert ipython_display.send_error.call_count == 1


def test_get_session_settings():
    assert magic.get_session_settings("something", False) == "something"
    assert magic.get_session_settings("something    ", False) == "something"
    assert magic.get_session_settings("    something", False) == "something"
    assert magic.get_session_settings("-f something", True) == "something"
    assert magic.get_session_settings("something -f", True) == "something"
    assert magic.get_session_settings("something", True) is None


def test_send_to_spark_with_non_empty_cell_error():
    line = "-i input -n name -t str"
    cell = "non empty"

    magic.session_started = True
    magic._do_send_to_spark = MagicMock()

    magic.send_to_spark(line, cell)

    assert ipython_display.send_error.call_count == 1


def test_send_to_spark_with_no_i_param_error():
    line = "-n name -t str"
    cell = ""

    magic.session_started = True
    magic._do_send_to_spark = MagicMock()

    magic.send_to_spark(line, cell)

    assert ipython_display.send_error.call_count == 1


def test_send_to_spark_ok():
    line = "-i input -n name -t str"
    cell = ""
    magic.shell.user_ns["input"] = None
    spark_controller.run_command = MagicMock(return_value=(True, line, "text/plain"))

    magic.send_to_spark(line, cell)

    assert ipython_display.write.called
    spark_controller.add_session.assert_called_once_with(
        magic.session_name,
        magic.endpoint,
        False,
        {"kind": constants.SESSION_KIND_PYSPARK},
    )
    spark_controller.run_command.assert_called_once_with(Command(cell), None)


def test_spark():
    line = ""
    cell = "some spark code"
    spark_controller.run_command = MagicMock(
        return_value=(True, line, constants.MIMETYPE_TEXT_PLAIN)
    )

    magic.spark(line, cell)

    ipython_display.write.assert_called_once_with(line)
    spark_controller.add_session.assert_called_once_with(
        magic.session_name,
        magic.endpoint,
        False,
        {"kind": constants.SESSION_KIND_PYSPARK},
    )
    spark_controller.run_command.assert_called_once_with(Command(cell), None)


def test_spark_with_argument():
    line = "-z"
    cell = "some spark code"
    spark_controller.run_command = MagicMock(return_value=(True, line))

    magic.spark(line, cell)

    assert ipython_display.send_error.call_count == 1


def test_spark_error():
    line = ""
    cell = "some spark code"
    spark_controller.run_command = MagicMock(
        return_value=(False, line, constants.MIMETYPE_TEXT_PLAIN)
    )

    magic.spark(line, cell)

    ipython_display.send_error.assert_called_once_with(
        constants.EXPECTED_ERROR_MSG.format(line)
    )
    spark_controller.add_session.assert_called_once_with(
        magic.session_name,
        magic.endpoint,
        False,
        {"kind": constants.SESSION_KIND_PYSPARK},
    )
    spark_controller.run_command.assert_called_once_with(Command(cell), None)


def test_spark_failed_session_start():
    line = ""
    cell = "some spark code"
    magic._do_not_call_start_session = MagicMock(return_value=False)

    ret = magic.spark(line, cell)

    assert ret is None
    assert ipython_display.write.call_count == 0
    assert spark_controller.add_session.call_count == 0
    assert spark_controller.run_command.call_count == 0


def test_spark_unexpected_exception():
    line = ""
    cell = "some spark code"
    spark_controller.run_command = MagicMock(side_effect=Exception("oups"))

    magic.spark(line, cell)
    spark_controller.run_command.assert_called_once_with(Command(cell), None)
    ipython_display.send_error.assert_called_once_with(
        constants.INTERNAL_ERROR_MSG.format(spark_controller.run_command.side_effect)
    )


def test_spark_expected_exception():
    line = ""
    cell = "some spark code"
    spark_controller.run_command = MagicMock(
        side_effect=SessionManagementException("oups")
    )

    magic.spark(line, cell)
    spark_controller.run_command.assert_called_once_with(Command(cell), None)
    ipython_display.send_error.assert_called_once_with(
        constants.EXPECTED_ERROR_MSG.format(spark_controller.run_command.side_effect)
    )


def test_spark_fatal_spark_statement_exception():
    with pytest.raises(SparkStatementException):
        conf.override_all(
            {
                "all_errors_are_fatal": True,
            }
        )

        line = ""
        cell = "some spark code"

        spark_controller.run_command = MagicMock(
            side_effect=SparkStatementException("Oh no!")
        )

        magic.spark(line, cell)


def test_spark_unexpected_exception_in_storing():
    line = "-o var_name"
    cell = "some spark code"
    side_effect = [(True, "ok", constants.MIMETYPE_TEXT_PLAIN), Exception("oups")]
    spark_controller.run_command = MagicMock(side_effect=side_effect)

    magic.spark(line, cell)
    assert spark_controller.run_command.call_count == 2
    spark_controller.run_command.assert_any_call(Command(cell), None)
    ipython_display.send_error.assert_called_with(
        constants.INTERNAL_ERROR_MSG.format(side_effect[1])
    )


def test_spark_expected_exception_in_storing():
    line = "-o var_name"
    cell = "some spark code"
    side_effect = [
        (True, "ok", constants.MIMETYPE_TEXT_PLAIN),
        SessionManagementException("oups"),
    ]
    spark_controller.run_command = MagicMock(side_effect=side_effect)

    magic.spark(line, cell)
    assert spark_controller.run_command.call_count == 2
    spark_controller.run_command.assert_any_call(Command(cell), None)
    ipython_display.send_error.assert_called_with(
        constants.EXPECTED_ERROR_MSG.format(side_effect[1])
    )


def test_spark_sample_options():
    line = "-o var_name -m sample -n 142 -r 0.3 -c True"
    cell = ""
    magic.execute_spark = MagicMock()
    ret = magic.spark(line, cell)

    magic.execute_spark.assert_called_once_with(
        cell, "var_name", "sample", 142, 0.3, None, True
    )


def test_spark_false_coerce():
    line = "-o var_name -m sample -n 142 -r 0.3 -c False"
    cell = ""
    magic.execute_spark = MagicMock()
    ret = magic.spark(line, cell)

    magic.execute_spark.assert_called_once_with(
        cell, "var_name", "sample", 142, 0.3, None, False
    )


def test_sql_without_output():
    line = ""
    cell = "some spark code"
    magic.execute_sqlquery = MagicMock()

    magic.sql(line, cell)

    spark_controller.add_session.assert_called_once_with(
        magic.session_name,
        magic.endpoint,
        False,
        {"kind": constants.SESSION_KIND_PYSPARK},
    )
    magic.execute_sqlquery.assert_called_once_with(
        cell, None, None, None, None, None, False, None
    )


def test_sql_with_output():
    line = "-o my_var"
    cell = "some spark code"
    magic.execute_sqlquery = MagicMock()

    magic.sql(line, cell)

    spark_controller.add_session.assert_called_once_with(
        magic.session_name,
        magic.endpoint,
        False,
        {"kind": constants.SESSION_KIND_PYSPARK},
    )
    magic.execute_sqlquery.assert_called_once_with(
        cell, None, None, None, None, "my_var", False, None
    )


def test_sql_exception():
    line = "-o my_var"
    cell = "some spark code"
    magic.execute_sqlquery = MagicMock(side_effect=ValueError("HAHAHAHAH"))

    magic.sql(line, cell)
    magic.execute_sqlquery.assert_called_once_with(
        cell, None, None, None, None, "my_var", False, None
    )
    ipython_display.send_error.assert_called_once_with(
        constants.INTERNAL_ERROR_MSG.format(magic.execute_sqlquery.side_effect)
    )


def test_sql_expected_exception():
    line = "-o my_var"
    cell = "some spark code"
    magic.execute_sqlquery = MagicMock(side_effect=HttpClientException("HAHAHAHAH"))

    magic.sql(line, cell)
    magic.execute_sqlquery.assert_called_once_with(
        cell, None, None, None, None, "my_var", False, None
    )
    ipython_display.send_error.assert_called_once_with(
        constants.EXPECTED_ERROR_MSG.format(magic.execute_sqlquery.side_effect)
    )


def test_sql_failed_session_start():
    line = ""
    cell = "some spark code"
    magic._do_not_call_start_session = MagicMock(return_value=False)

    ret = magic.sql(line, cell)

    assert ret is None
    assert spark_controller.add_session.call_count == 0
    assert spark_controller.execute_sqlquery.call_count == 0


def test_sql_quiet():
    line = "-q -o Output"
    cell = ""
    magic.execute_sqlquery = MagicMock()

    ret = magic.sql(line, cell)

    spark_controller.add_session.assert_called_once_with(
        magic.session_name,
        magic.endpoint,
        False,
        {"kind": constants.SESSION_KIND_PYSPARK},
    )
    magic.execute_sqlquery.assert_called_once_with(
        cell, None, None, None, None, "Output", True, None
    )


def test_sql_sample_options():
    line = "-q -m sample -n 142 -r 0.3 -c True"
    cell = ""
    magic.execute_sqlquery = MagicMock()

    ret = magic.sql(line, cell)

    spark_controller.add_session.assert_called_once_with(
        magic.session_name,
        magic.endpoint,
        False,
        {"kind": constants.SESSION_KIND_PYSPARK},
    )
    magic.execute_sqlquery.assert_called_once_with(
        cell, "sample", 142, 0.3, None, None, True, True
    )


def test_sql_false_coerce():
    line = "-q -m sample -n 142 -r 0.3 -c False"
    cell = ""
    magic.execute_sqlquery = MagicMock()

    ret = magic.sql(line, cell)

    spark_controller.add_session.assert_called_once_with(
        magic.session_name,
        magic.endpoint,
        False,
        {"kind": constants.SESSION_KIND_PYSPARK},
    )
    magic.execute_sqlquery.assert_called_once_with(
        cell, "sample", 142, 0.3, None, None, True, False
    )


def test_cleanup_without_force():
    line = ""
    cell = ""
    magic.session_started = True
    spark_controller.cleanup_endpoint = MagicMock()
    spark_controller.delete_session_by_name = MagicMock()

    magic.cleanup(line, cell)

    _assert_magic_successful_event_emitted_once("cleanup")

    assert ipython_display.send_error.call_count == 1
    assert spark_controller.cleanup_endpoint.call_count == 0


def test_cleanup_with_force():
    line = "-f"
    cell = ""
    magic.session_started = True
    spark_controller.cleanup_endpoint = MagicMock()
    spark_controller.delete_session_by_name = MagicMock()

    magic.cleanup(line, cell)

    _assert_magic_successful_event_emitted_once("cleanup")

    spark_controller.cleanup_endpoint.assert_called_once_with(magic.endpoint)
    spark_controller.delete_session_by_name.assert_called_once_with(magic.session_name)


def test_cleanup_with_cell_content():
    line = "-f"
    cell = "HEHEHE"
    msg = "Cell body for %%cleanup magic must be empty; got 'HEHEHE' instead"
    magic.session_started = True
    spark_controller.cleanup_endpoint = MagicMock()
    spark_controller.delete_session_by_name = MagicMock()

    magic.cleanup(line, cell)
    assert ipython_display.send_error.call_count == 1
    _assert_magic_failure_event_emitted_once("cleanup", BadUserDataException(msg))


def test_cleanup_exception():
    line = "-f"
    cell = ""
    magic.session_started = True
    spark_controller.cleanup_endpoint = MagicMock(
        side_effect=ArithmeticError("DIVISION BY ZERO OH NO")
    )

    magic.cleanup(line, cell)
    _assert_magic_failure_event_emitted_once(
        "cleanup", spark_controller.cleanup_endpoint.side_effect
    )
    spark_controller.cleanup_endpoint.assert_called_once_with(magic.endpoint)
    ipython_display.send_error.assert_called_once_with(
        constants.INTERNAL_ERROR_MSG.format(
            spark_controller.cleanup_endpoint.side_effect
        )
    )


def test_delete_without_force():
    session_id = 0
    line = "-s {}".format(session_id)
    cell = ""
    spark_controller.delete_session_by_id = MagicMock()
    spark_controller.get_session_id_for_client = MagicMock(return_value=session_id)

    magic.delete(line, cell)

    _assert_magic_successful_event_emitted_once("delete")

    assert ipython_display.send_error.call_count == 1
    assert spark_controller.delete_session_by_id.call_count == 0


def test_delete_without_session_id():
    session_id = 0
    line = ""
    cell = ""
    spark_controller.delete_session_by_id = MagicMock()
    spark_controller.get_session_id_for_client = MagicMock(return_value=session_id)

    magic.delete(line, cell)

    _assert_magic_successful_event_emitted_once("delete")

    assert ipython_display.send_error.call_count == 1
    assert spark_controller.delete_session_by_id.call_count == 0


def test_delete_with_force_same_session():
    session_id = 0
    line = "-f -s {}".format(session_id)
    cell = ""
    spark_controller.delete_session_by_id = MagicMock()
    spark_controller.get_session_id_for_client = MagicMock(return_value=session_id)

    magic.delete(line, cell)

    _assert_magic_successful_event_emitted_once("delete")

    assert ipython_display.send_error.call_count == 1
    assert spark_controller.delete_session_by_id.call_count == 0


def test_delete_with_force_none_session():
    # This happens when session has not been created
    session_id = 0
    line = "-f -s {}".format(session_id)
    cell = ""
    spark_controller.delete_session_by_id = MagicMock()
    spark_controller.get_session_id_for_client = MagicMock(return_value=None)

    magic.delete(line, cell)

    _assert_magic_successful_event_emitted_once("delete")

    spark_controller.get_session_id_for_client.assert_called_once_with(
        magic.session_name
    )
    spark_controller.delete_session_by_id.assert_called_once_with(
        magic.endpoint, session_id
    )


def test_delete_with_cell_content():
    # This happens when session has not been created
    session_id = 0
    line = "-f -s {}".format(session_id)
    cell = "~~~"
    msg = "Cell body for %%delete magic must be empty; got '~~~' instead"
    spark_controller.delete_session_by_id = MagicMock()
    spark_controller.get_session_id_for_client = MagicMock(return_value=None)

    magic.delete(line, cell)

    _assert_magic_failure_event_emitted_once("delete", BadUserDataException(msg))
    assert ipython_display.send_error.call_count == 1


def test_delete_with_force_different_session():
    # This happens when session has not been created
    session_id = 0
    line = "-f -s {}".format(session_id)
    cell = ""
    spark_controller.delete_session_by_id = MagicMock()
    spark_controller.get_session_id_for_client = MagicMock(return_value="1")

    magic.delete(line, cell)

    _assert_magic_successful_event_emitted_once("delete")

    spark_controller.get_session_id_for_client.assert_called_once_with(
        magic.session_name
    )
    spark_controller.delete_session_by_id.assert_called_once_with(
        magic.endpoint, session_id
    )


def test_delete_exception():
    # This happens when session has not been created
    session_id = 0
    line = "-f -s {}".format(session_id)
    cell = ""
    spark_controller.delete_session_by_id = MagicMock(
        side_effect=DataFrameParseException("wow")
    )
    spark_controller.get_session_id_for_client = MagicMock()

    magic.delete(line, cell)
    _assert_magic_failure_event_emitted_once(
        "delete", spark_controller.delete_session_by_id.side_effect
    )
    spark_controller.get_session_id_for_client.assert_called_once_with(
        magic.session_name
    )
    spark_controller.delete_session_by_id.assert_called_once_with(
        magic.endpoint, session_id
    )
    ipython_display.send_error.assert_called_once_with(
        constants.INTERNAL_ERROR_MSG.format(
            spark_controller.delete_session_by_id.side_effect
        )
    )


def test_delete_expected_exception():
    # This happens when session has not been created
    session_id = 0
    line = "-f -s {}".format(session_id)
    cell = ""
    spark_controller.delete_session_by_id = MagicMock(
        side_effect=LivyClientTimeoutException("wow")
    )
    spark_controller.get_session_id_for_client = MagicMock()

    magic.delete(line, cell)
    _assert_magic_failure_event_emitted_once(
        "delete", spark_controller.delete_session_by_id.side_effect
    )
    spark_controller.get_session_id_for_client.assert_called_once_with(
        magic.session_name
    )
    spark_controller.delete_session_by_id.assert_called_once_with(
        magic.endpoint, session_id
    )
    ipython_display.send_error.assert_called_once_with(
        constants.EXPECTED_ERROR_MSG.format(
            spark_controller.delete_session_by_id.side_effect
        )
    )


def test_start_session_displays_fatal_error_when_session_throws():
    e = ValueError("Failed to create the SqlContext.\nError, '{}'".format("Exception"))
    magic.spark_controller.add_session = MagicMock(side_effect=e)
    magic.language = constants.LANG_SCALA
    magic.ipython_display = ipython_display
    magic.ipython_display.send_error = MagicMock()

    magic._do_not_call_start_session("")

    magic.spark_controller.add_session.assert_called_once_with(
        magic.session_name,
        magic.endpoint,
        False,
        {"kind": constants.SESSION_KIND_SPARK},
    )
    assert magic.fatal_error
    assert magic.fatal_error_message == conf.fatal_error_suggestion().format(str(e))


def test_start_session_when_retry_fatal_error_is_not_allowed_by_default():
    e = ValueError("Failed to create the SqlContext.\nError, '{}'".format("Exception"))
    magic.spark_controller.add_session = MagicMock(side_effect=e)
    magic.language = constants.LANG_SCALA
    magic.ipython_display = ipython_display
    magic.ipython_display.send_error = MagicMock()

    # first session creation
    magic._do_not_call_start_session("")
    # retry session creation
    magic._do_not_call_start_session("")

    # call add_session once and call send_error twice to show the error message
    magic.spark_controller.add_session.assert_called_once_with(
        magic.session_name,
        magic.endpoint,
        False,
        {"kind": constants.SESSION_KIND_SPARK},
    )
    assert magic.ipython_display.send_error.call_count == 2


def test_retry_start_session_when_retry_fatal_error_is_allowed():
    e = ValueError("Failed to create the SqlContext.\nError, '{}'".format("Exception"))
    magic.spark_controller.add_session = MagicMock(side_effect=e)
    magic.language = constants.LANG_SCALA
    magic.ipython_display = ipython_display
    magic.ipython_display.send_error = MagicMock()
    magic.allow_retry_fatal = True

    # first session creation - failed
    session_created = magic._do_not_call_start_session("")
    magic.spark_controller.add_session.assert_called_once_with(
        magic.session_name,
        magic.endpoint,
        False,
        {"kind": constants.SESSION_KIND_SPARK},
    )
    assert not session_created
    assert magic.fatal_error
    assert magic.fatal_error_message == conf.fatal_error_suggestion().format(str(e))

    # retry session creation - successful
    magic.spark_controller.add_session = MagicMock()
    session_created = magic._do_not_call_start_session("")
    magic.spark_controller.add_session.assert_called_once_with(
        magic.session_name,
        magic.endpoint,
        False,
        {"kind": constants.SESSION_KIND_SPARK},
    )
    print(session_created)
    assert session_created
    assert magic.session_started
    assert not magic.fatal_error
    assert magic.fatal_error_message == ""

    # show error message only once
    assert magic.ipython_display.send_error.call_count == 1


def test_allow_retry_fatal():
    assert not magic.allow_retry_fatal
    magic._do_not_call_allow_retry_fatal("")
    assert magic.allow_retry_fatal


def test_kernel_magics_names():
    """The magics machinery in IPython depends on the docstrings and method
    names matching up correctly."""
    assert magic.help.__name__ == "help"
    assert magic.local.__name__ == "local"
    assert magic.info.__name__ == "info"
    assert magic.logs.__name__ == "logs"
    assert magic.configure.__name__ == "configure"
    assert magic.spark.__name__ == "spark"
    assert magic.sql.__name__ == "sql"
    assert magic.cleanup.__name__ == "cleanup"
    assert magic.delete.__name__ == "delete"


def _assert_magic_successful_event_emitted_once(name):
    magic._generate_uuid.assert_called_once_with()
    spark_events.emit_magic_execution_start_event.assert_called_once_with(
        name, constants.SESSION_KIND_PYSPARK, magic._generate_uuid.return_value
    )
    spark_events.emit_magic_execution_end_event.assert_called_once_with(
        name,
        constants.SESSION_KIND_PYSPARK,
        magic._generate_uuid.return_value,
        True,
        "",
        "",
    )


def _assert_magic_failure_event_emitted_once(name, error):
    magic._generate_uuid.assert_called_once_with()
    spark_events.emit_magic_execution_start_event.assert_called_once_with(
        name, constants.SESSION_KIND_PYSPARK, magic._generate_uuid.return_value
    )
    spark_events.emit_magic_execution_end_event.assert_called_once_with(
        name,
        constants.SESSION_KIND_PYSPARK,
        magic._generate_uuid.return_value,
        False,
        error.__class__.__name__,
        str(error),
    )



import json

try:
    from jupyter_server.utils import url_path_join
    from jupyter_server.base.handlers import JupyterHandler as BaseHandler
except ImportError:
    from notebook.utils import url_path_join
    from notebook.base.handlers import IPythonHandler as BaseHandler

from tornado import web
from tornado import gen
from tornado.web import MissingArgumentError
from tornado.escape import json_decode

from sparkmagic.kernels.kernelmagics import KernelMagics
import sparkmagic.utils.configuration as conf
from sparkmagic.utils import constants
from sparkmagic.utils.sparkevents import SparkEvents
from sparkmagic.utils.sparklogger import SparkLog


class ReconnectHandler(BaseHandler):
    logger = None

    @web.authenticated
    @gen.coroutine
    def post(self):
        self.logger = SparkLog("ReconnectHandler")

        spark_events = self._get_spark_events()

        try:
            data = json_decode(self.request.body)
        except ValueError as e:
            self.set_status(400)
            msg = "Invalid JSON in request body."
            self.logger.error(msg)
            self.finish(msg)
            spark_events.emit_cluster_change_event(None, 400, False, msg)
            return

        endpoint = None
        try:
            path = self._get_argument_or_raise(data, "path")
            username = self._get_argument_or_raise(data, "username")
            password = self._get_argument_or_raise(data, "password")
            endpoint = self._get_argument_or_raise(data, "endpoint")
            auth = self._get_argument_if_exists(data, "auth")
            if auth is None:
                if username == "" and password == "":
                    auth = constants.NO_AUTH
                else:
                    auth = constants.AUTH_BASIC
        except MissingArgumentError as e:
            self.set_status(400)
            self.finish(str(e))
            self.logger.error(str(e))
            spark_events.emit_cluster_change_event(endpoint, 400, False, str(e))
            return

        kernel_name = self._get_kernel_name(data)

        # Get kernel manager, create a new kernel if none exists or restart the existing one when applicable
        kernel_manager = yield self._get_kernel_manager(path, kernel_name)

        # Execute code
        client = kernel_manager.client()
        code = "%{} -s {} -u {} -p {} -t {}".format(
            KernelMagics._do_not_call_change_endpoint.__name__,
            endpoint,
            username,
            password,
            auth,
        )
        response_id = client.execute(code, silent=False, store_history=False)
        msg = client.get_shell_msg(response_id)

        # Get execution info
        successful_message = self._msg_successful(msg)
        error = self._msg_error(msg)
        if successful_message:
            status_code = 200
        else:
            self.logger.error("Code to reconnect errored out: {}".format(error))
            status_code = 500

        # Post execution info
        self.set_status(status_code)
        self.finish(
            json.dumps(dict(success=successful_message, error=error), sort_keys=True)
        )
        spark_events.emit_cluster_change_event(
            endpoint, status_code, successful_message, error
        )

    def _get_kernel_name(self, data):
        kernel_name = self._get_argument_if_exists(data, "kernelname")
        self.logger.debug("Kernel name is {}".format(kernel_name))
        if kernel_name is None:
            kernel_name = conf.server_extension_default_kernel_name()
            self.logger.debug("Defaulting to kernel name {}".format(kernel_name))
        return kernel_name

    def _get_argument_if_exists(self, data, key):
        return data.get(key)

    def _get_argument_or_raise(self, data, key):
        try:
            return data[key]
        except KeyError:
            raise MissingArgumentError(key)

    @gen.coroutine
    def _get_kernel_manager(self, path, kernel_name):
        sessions = self.session_manager.list_sessions()

        kernel_id = None
        for session in sessions:
            if session["notebook"]["path"] == path:
                session_id = session["id"]
                kernel_id = session["kernel"]["id"]
                existing_kernel_name = session["kernel"]["name"]
                break

        if kernel_id is None:
            self.logger.debug("Kernel not found. Starting a new kernel.")
            k_m = yield self._get_kernel_manager_new_session(path, kernel_name)
        elif existing_kernel_name != kernel_name:
            self.logger.debug(
                "Existing kernel name '{}' does not match requested '{}'. Starting a new kernel.".format(
                    existing_kernel_name, kernel_name
                )
            )
            self._delete_session(session_id)
            k_m = yield self._get_kernel_manager_new_session(path, kernel_name)
        else:
            self.logger.debug("Kernel found. Restarting kernel.")
            k_m = self.kernel_manager.get_kernel(kernel_id)
            k_m.restart_kernel()

        raise gen.Return(k_m)

    @gen.coroutine
    def _get_kernel_manager_new_session(self, path, kernel_name):
        model_future = self.session_manager.create_session(
            kernel_name=kernel_name, path=path, type="notebook"
        )
        model = yield model_future
        kernel_id = model["kernel"]["id"]
        self.logger.debug("Kernel created with id {}".format(str(kernel_id)))
        k_m = self.kernel_manager.get_kernel(kernel_id)
        raise gen.Return(k_m)

    def _delete_session(self, session_id):
        self.session_manager.delete_session(session_id)

    def _msg_status(self, msg):
        return msg["content"]["status"]

    def _msg_successful(self, msg):
        return self._msg_status(msg) == "ok"

    def _msg_error(self, msg):
        if self._msg_status(msg) != "error":
            return None
        return "{}:\n{}".format(msg["content"]["ename"], msg["content"]["evalue"])

    def _get_spark_events(self):
        spark_events = getattr(self, "spark_events", None)
        if spark_events is None:
            return SparkEvents()
        return spark_events


def load_jupyter_server_extension(nb_app):
    nb_app.log.info("sparkmagic extension enabled!")
    web_app = nb_app.web_app

    base_url = web_app.settings["base_url"]
    host_pattern = ".*$"

    route_pattern_reconnect = url_path_join(base_url, "/reconnectsparkmagic")
    handlers = [(route_pattern_reconnect, ReconnectHandler)]

    web_app.add_handlers(host_pattern, handlers)



from mock import MagicMock

import sparkmagic.utils.configuration as conf
from sparkmagic.utils.utils import parse_argstring_or_throw, initialize_auth
from sparkmagic.utils.constants import (
    EXPECTED_ERROR_MSG,
    MIMETYPE_TEXT_PLAIN,
    NO_AUTH,
    AUTH_BASIC,
)
from sparkmagic.magics.remotesparkmagics import RemoteSparkMagics
from sparkmagic.livyclientlib.command import Command
from sparkmagic.livyclientlib.endpoint import Endpoint
from sparkmagic.livyclientlib.exceptions import *
from sparkmagic.livyclientlib.sqlquery import SQLQuery
from sparkmagic.livyclientlib.sparkstorecommand import SparkStoreCommand

magic = None
spark_controller = None
shell = None
ipython_display = None


def setup_function():
    global magic, spark_controller, shell, ipython_display
    conf.override_all({})

    magic = RemoteSparkMagics(shell=None, widget=MagicMock())
    magic.shell = shell = MagicMock()
    magic.ipython_display = ipython_display = MagicMock()
    magic.spark_controller = spark_controller = MagicMock()


def teardown_function():
    pass


def test_info_command_parses():
    print_info_mock = MagicMock()
    magic._print_local_info = print_info_mock
    command = "info"

    magic.spark(command)

    print_info_mock.assert_called_once_with()


def test_info_endpoint_command_parses():
    print_info_mock = MagicMock()
    magic._print_endpoint_info = print_info_mock
    command = "info -u http://microsoft.com -i 1234"
    spark_controller.get_all_sessions_endpoint_info = MagicMock(return_value=None)

    magic.spark(command)

    print_info_mock.assert_called_once_with(None, 1234)


def test_info_command_exception():
    print_info_mock = MagicMock(side_effect=LivyClientTimeoutException("OHHHHHOHOHOHO"))
    magic._print_local_info = print_info_mock
    command = "info"

    magic.spark(command)

    print_info_mock.assert_called_once_with()
    ipython_display.send_error.assert_called_once_with(
        EXPECTED_ERROR_MSG.format(print_info_mock.side_effect)
    )


def test_add_sessions_command_parses():
    # Do not skip and python
    add_sessions_mock = MagicMock()
    spark_controller.add_session = add_sessions_mock
    command = "add"
    name = "-s name"
    language = "-l python"
    connection_string = "-u http://url.com -t {} -a sdf -p w".format(AUTH_BASIC)
    line = " ".join([command, name, language, connection_string])

    magic.spark(line)
    args = parse_argstring_or_throw(RemoteSparkMagics.spark, line)
    add_sessions_mock.assert_called_once_with(
        "name",
        Endpoint("http://url.com", initialize_auth(args)),
        False,
        {"kind": "pyspark"},
    )
    # Skip and scala - upper case
    add_sessions_mock = MagicMock()
    spark_controller.add_session = add_sessions_mock
    command = "add"
    name = "-s name"
    language = "-l scala"
    connection_string = "--url http://location:port"
    line = " ".join([command, name, language, connection_string, "-k"])

    magic.spark(line)
    args = parse_argstring_or_throw(RemoteSparkMagics.spark, line)
    args.auth = NO_AUTH
    add_sessions_mock.assert_called_once_with(
        "name",
        Endpoint("http://location:port", initialize_auth(args)),
        True,
        {"kind": "spark"},
    )


def test_add_sessions_command_parses_kerberos():
    # Do not skip and python
    add_sessions_mock = MagicMock()
    spark_controller.add_session = add_sessions_mock
    command = "add"
    name = "-s name"
    language = "-l python"
    connection_string = "-u http://url.com -t {}".format("Kerberos")
    line = " ".join([command, name, language, connection_string])
    magic.spark(line)
    args = parse_argstring_or_throw(RemoteSparkMagics.spark, line)
    auth_instance = initialize_auth(args)

    add_sessions_mock.assert_called_once_with(
        "name",
        Endpoint("http://url.com", initialize_auth(args)),
        False,
        {"kind": "pyspark"},
    )
    assert auth_instance.url == "http://url.com"


def test_add_sessions_command_exception():
    # Do not skip and python
    add_sessions_mock = MagicMock(side_effect=BadUserDataException("hehe"))
    spark_controller.add_session = add_sessions_mock
    command = "add"
    name = "-s name"
    language = "-l python"
    connection_string = "-u http://url.com -t {} -a sdf -p w".format(AUTH_BASIC)
    line = " ".join([command, name, language, connection_string])

    magic.spark(line)
    args = parse_argstring_or_throw(RemoteSparkMagics.spark, line)
    add_sessions_mock.assert_called_once_with(
        "name",
        Endpoint("http://url.com", initialize_auth(args)),
        False,
        {"kind": "pyspark"},
    )
    ipython_display.send_error.assert_called_once_with(
        EXPECTED_ERROR_MSG.format(add_sessions_mock.side_effect)
    )


def test_add_sessions_command_extra_properties():
    conf.override_all({})
    magic.spark("config", '{"extra": "yes"}')
    assert conf.session_configs() == {"extra": "yes"}

    add_sessions_mock = MagicMock()
    spark_controller.add_session = add_sessions_mock
    command = "add"
    name = "-s name"
    language = "-l scala"
    connection_string = "-u http://livyendpoint.com"
    line = " ".join([command, name, language, connection_string])

    magic.spark(line)
    args = parse_argstring_or_throw(RemoteSparkMagics.spark, line)
    args.auth = NO_AUTH
    add_sessions_mock.assert_called_once_with(
        "name",
        Endpoint("http://livyendpoint.com", initialize_auth(args)),
        False,
        {"kind": "spark", "extra": "yes"},
    )
    conf.override_all({})


def test_delete_sessions_command_parses():
    mock_method = MagicMock()
    spark_controller.delete_session_by_name = mock_method
    command = "delete -s name"
    magic.spark(command)
    mock_method.assert_called_once_with("name")

    command = "delete -u URL -t {} -a username -p password -i 4".format(AUTH_BASIC)
    mock_method = MagicMock()
    spark_controller.delete_session_by_id = mock_method
    magic.spark(command)
    args = parse_argstring_or_throw(RemoteSparkMagics.spark, command)
    mock_method.assert_called_once_with(Endpoint("URL", initialize_auth(args)), 4)


def test_delete_sessions_command_exception():
    mock_method = MagicMock(side_effect=LivyUnexpectedStatusException("FEEEEEELINGS"))
    spark_controller.delete_session_by_name = mock_method
    command = "delete -s name"
    magic.spark(command)
    mock_method.assert_called_once_with("name")
    ipython_display.send_error.assert_called_once_with(
        EXPECTED_ERROR_MSG.format(mock_method.side_effect)
    )


def test_cleanup_command_parses():
    mock_method = MagicMock()
    spark_controller.cleanup = mock_method
    line = "cleanup"

    magic.spark(line)

    mock_method.assert_called_once_with()


def test_cleanup_command_exception():
    mock_method = MagicMock(
        side_effect=SessionManagementException("Livy did something VERY BAD")
    )
    spark_controller.cleanup = mock_method
    line = "cleanup"

    magic.spark(line)
    mock_method.assert_called_once_with()
    ipython_display.send_error.assert_called_once_with(
        EXPECTED_ERROR_MSG.format(mock_method.side_effect)
    )


def test_cleanup_endpoint_command_parses():
    mock_method = MagicMock()
    spark_controller.cleanup_endpoint = mock_method
    line = "cleanup -u endp"

    magic.spark(line)
    args = parse_argstring_or_throw(RemoteSparkMagics.spark, line)
    args.auth = NO_AUTH
    mock_method.assert_called_once_with(Endpoint("endp", initialize_auth(args)))

    line = "cleanup -u endp -a user -p passw -t {}".format(AUTH_BASIC)
    magic.spark(line)
    args = parse_argstring_or_throw(RemoteSparkMagics.spark, line)
    mock_method.assert_called_with(Endpoint("endp", initialize_auth(args)))


def test_bad_command_writes_error():
    line = "bad_command"
    usage = "Please look at usage of %spark by executing `%spark?`."

    magic.spark(line)

    ipython_display.send_error.assert_called_once_with(
        "Subcommand '{}' not found. {}".format(line, usage)
    )


def test_run_cell_command_parses():
    run_cell_method = MagicMock()
    result_value = ""
    run_cell_method.return_value = (True, result_value, MIMETYPE_TEXT_PLAIN)
    spark_controller.run_command = run_cell_method

    command = "-s"
    name = "sessions_name"
    line = " ".join([command, name])
    cell = "cell code"

    result = magic.spark(line, cell)

    run_cell_method.assert_called_once_with(Command(cell), name)
    assert result is None
    ipython_display.write.assert_called_once_with(result_value)


def test_run_cell_command_writes_to_err():
    run_cell_method = MagicMock()
    result_value = ""
    run_cell_method.return_value = (False, result_value, MIMETYPE_TEXT_PLAIN)
    spark_controller.run_command = run_cell_method

    command = "-s"
    name = "sessions_name"
    line = " ".join([command, name])
    cell = "cell code"

    result = magic.spark(line, cell)

    run_cell_method.assert_called_once_with(Command(cell), name)
    assert result is None
    ipython_display.send_error.assert_called_once_with(
        EXPECTED_ERROR_MSG.format(result_value)
    )


def test_run_cell_command_exception():
    run_cell_method = MagicMock()
    run_cell_method.side_effect = HttpClientException("meh")
    spark_controller.run_command = run_cell_method

    command = "-s"
    name = "sessions_name"
    line = " ".join([command, name])
    cell = "cell code"

    result = magic.spark(line, cell)

    run_cell_method.assert_called_once_with(Command(cell), name)
    assert result is None
    ipython_display.send_error.assert_called_once_with(
        EXPECTED_ERROR_MSG.format(run_cell_method.side_effect)
    )


def test_run_spark_command_parses():
    magic.execute_spark = MagicMock()

    command = "-s"
    name = "sessions_name"
    context = "-c"
    context_name = "spark"
    meth = "-m"
    method_name = "sample"
    line = " ".join([command, name, context, context_name, meth, method_name])
    cell = "cell code"

    result = magic.spark(line, cell)

    magic.execute_spark.assert_called_once_with(
        "cell code", None, "sample", None, None, "sessions_name", None
    )


def test_run_spark_command_parses_with_coerce():
    magic.execute_spark = MagicMock()

    command = "-s"
    name = "sessions_name"
    context = "-c"
    context_name = "spark"
    meth = "-m"
    method_name = "sample"
    coer = "--coerce"
    coerce_value = "True"
    line = " ".join(
        [command, name, context, context_name, meth, method_name, coer, coerce_value]
    )
    cell = "cell code"

    result = magic.spark(line, cell)

    magic.execute_spark.assert_called_once_with(
        "cell code", None, "sample", None, None, "sessions_name", True
    )


def test_run_spark_command_parses_with_coerce_false():
    magic.execute_spark = MagicMock()

    command = "-s"
    name = "sessions_name"
    context = "-c"
    context_name = "spark"
    meth = "-m"
    method_name = "sample"
    coer = "--coerce"
    coerce_value = "False"
    line = " ".join(
        [command, name, context, context_name, meth, method_name, coer, coerce_value]
    )
    cell = "cell code"

    result = magic.spark(line, cell)

    magic.execute_spark.assert_called_once_with(
        "cell code", None, "sample", None, None, "sessions_name", False
    )


def test_run_sql_command_parses_with_coerce_false():
    magic.execute_sqlquery = MagicMock()

    command = "-s"
    name = "sessions_name"
    context = "-c"
    context_name = "sql"
    meth = "-m"
    method_name = "sample"
    coer = "--coerce"
    coerce_value = "False"
    line = " ".join(
        [command, name, context, context_name, meth, method_name, coer, coerce_value]
    )
    cell = "cell code"

    result = magic.spark(line, cell)

    magic.execute_sqlquery.assert_called_once_with(
        "cell code", "sample", None, None, "sessions_name", None, False, False
    )


def test_run_spark_with_store_command_parses():
    magic.execute_spark = MagicMock()

    command = "-s"
    name = "sessions_name"
    context = "-c"
    context_name = "spark"
    meth = "-m"
    method_name = "sample"
    output = "-o"
    output_var = "var_name"
    line = " ".join(
        [command, name, context, context_name, meth, method_name, output, output_var]
    )
    cell = "cell code"

    result = magic.spark(line, cell)
    magic.execute_spark.assert_called_once_with(
        "cell code", "var_name", "sample", None, None, "sessions_name", None
    )


def test_run_spark_with_store_correct_calls():
    run_cell_method = MagicMock()
    run_cell_method.return_value = (True, "", MIMETYPE_TEXT_PLAIN)
    spark_controller.run_command = run_cell_method

    command = "-s"
    name = "sessions_name"
    context = "-c"
    context_name = "spark"
    meth = "-m"
    method_name = "sample"
    output = "-o"
    output_var = "var_name"
    coer = "--coerce"
    coerce_value = "True"
    line = " ".join(
        [
            command,
            name,
            context,
            context_name,
            meth,
            method_name,
            output,
            output_var,
            coer,
            coerce_value,
        ]
    )
    cell = "cell code"

    result = magic.spark(line, cell)

    run_cell_method.assert_any_call(Command(cell), name)
    run_cell_method.assert_any_call(
        SparkStoreCommand(output_var, samplemethod=method_name, coerce=True), name
    )


def test_run_spark_command_exception():
    run_cell_method = MagicMock()
    run_cell_method.side_effect = LivyUnexpectedStatusException("WOW")
    spark_controller.run_command = run_cell_method

    command = "-s"
    name = "sessions_name"
    context = "-c"
    context_name = "spark"
    meth = "-m"
    method_name = "sample"
    output = "-o"
    output_var = "var_name"
    line = " ".join(
        [command, name, context, context_name, meth, method_name, output, output_var]
    )
    cell = "cell code"

    result = magic.spark(line, cell)

    run_cell_method.assert_any_call(Command(cell), name)
    ipython_display.send_error.assert_called_once_with(
        EXPECTED_ERROR_MSG.format(run_cell_method.side_effect)
    )


def test_run_spark_command_exception_while_storing():
    run_cell_method = MagicMock()
    exception = LivyUnexpectedStatusException("WOW")
    run_cell_method.side_effect = [(True, "", MIMETYPE_TEXT_PLAIN), exception]
    spark_controller.run_command = run_cell_method

    command = "-s"
    name = "sessions_name"
    context = "-c"
    context_name = "spark"
    meth = "-m"
    method_name = "sample"
    output = "-o"
    output_var = "var_name"
    line = " ".join(
        [command, name, context, context_name, meth, method_name, output, output_var]
    )
    cell = "cell code"

    result = magic.spark(line, cell)

    run_cell_method.assert_any_call(Command(cell), name)
    run_cell_method.assert_any_call(
        SparkStoreCommand(output_var, samplemethod=method_name), name
    )
    ipython_display.write.assert_called_once_with("")
    ipython_display.send_error.assert_called_once_with(
        EXPECTED_ERROR_MSG.format(exception)
    )


def test_run_sql_command_parses():
    run_cell_method = MagicMock()
    run_cell_method.return_value = (True, "", MIMETYPE_TEXT_PLAIN)
    spark_controller.run_sqlquery = run_cell_method

    command = "-s"
    name = "sessions_name"
    context = "-c"
    context_name = "sql"
    meth = "-m"
    method_name = "sample"
    line = " ".join([command, name, context, context_name, meth, method_name])
    cell = "cell code"

    result = magic.spark(line, cell)

    run_cell_method.assert_called_once_with(
        SQLQuery(cell, samplemethod=method_name), name
    )
    assert result is not None


def test_run_sql_command_exception():
    run_cell_method = MagicMock()
    run_cell_method.side_effect = LivyUnexpectedStatusException("WOW")
    spark_controller.run_sqlquery = run_cell_method

    command = "-s"
    name = "sessions_name"
    context = "-c"
    context_name = "sql"
    meth = "-m"
    method_name = "sample"
    line = " ".join([command, name, context, context_name, meth, method_name])
    cell = "cell code"

    result = magic.spark(line, cell)

    run_cell_method.assert_called_once_with(
        SQLQuery(cell, samplemethod=method_name), name
    )
    ipython_display.send_error.assert_called_once_with(
        EXPECTED_ERROR_MSG.format(run_cell_method.side_effect)
    )


def test_run_sql_command_knows_how_to_be_quiet():
    run_cell_method = MagicMock()
    run_cell_method.return_value = (True, "", MIMETYPE_TEXT_PLAIN)
    spark_controller.run_sqlquery = run_cell_method

    command = "-s"
    name = "sessions_name"
    context = "-c"
    context_name = "sql"
    quiet = "-q"
    meth = "-m"
    method_name = "sample"
    line = " ".join([command, name, context, context_name, quiet, meth, method_name])
    cell = "cell code"

    result = magic.spark(line, cell)

    run_cell_method.assert_called_once_with(
        SQLQuery(cell, samplemethod=method_name), name
    )
    assert result is None


def test_logs_subcommand():
    get_logs_method = MagicMock()
    result_value = ""
    get_logs_method.return_value = result_value
    spark_controller.get_logs = get_logs_method

    command = "logs -s"
    name = "sessions_name"
    line = " ".join([command, name])
    cell = "cell code"

    result = magic.spark(line, cell)

    get_logs_method.assert_called_once_with(name)
    assert result is None
    ipython_display.write.assert_called_once_with(result_value)


def test_logs_exception():
    get_logs_method = MagicMock(
        side_effect=LivyUnexpectedStatusException("How did this happen?")
    )
    result_value = ""
    get_logs_method.return_value = result_value
    spark_controller.get_logs = get_logs_method

    command = "logs -s"
    name = "sessions_name"
    line = " ".join([command, name])
    cell = "cell code"

    result = magic.spark(line, cell)

    get_logs_method.assert_called_once_with(name)
    assert result is None
    ipython_display.send_error.assert_called_once_with(
        EXPECTED_ERROR_MSG.format(get_logs_method.side_effect)
    )



from sparkmagic.livyclientlib.configurableretrypolicy import ConfigurableRetryPolicy
import sparkmagic.utils.configuration as conf
from sparkmagic.livyclientlib.exceptions import BadUserConfigurationException


def test_with_empty_list():
    times = []
    max_retries = 5
    policy = ConfigurableRetryPolicy(times, max_retries)

    assert 5 == policy.seconds_to_sleep(0)
    assert 5 == policy.seconds_to_sleep(4)
    assert 5 == policy.seconds_to_sleep(5)
    assert 5 == policy.seconds_to_sleep(6)

    # Check based on retry count
    assert True is policy.should_retry(500, False, 0)
    assert True is policy.should_retry(500, False, 4)
    assert True is policy.should_retry(500, False, 5)
    assert False is policy.should_retry(500, False, 6)

    # Check based on status code
    assert False is policy.should_retry(201, False, 0)
    assert False is policy.should_retry(201, False, 6)

    # Check based on error
    assert True is policy.should_retry(201, True, 0)
    assert True is policy.should_retry(201, True, 6)


def test_with_one_element_list():
    times = [2]
    max_retries = 5
    policy = ConfigurableRetryPolicy(times, max_retries)

    assert 2 == policy.seconds_to_sleep(0)
    assert 2 == policy.seconds_to_sleep(4)
    assert 2 == policy.seconds_to_sleep(5)
    assert 2 == policy.seconds_to_sleep(6)

    # Check based on retry count
    assert True is policy.should_retry(500, False, 0)
    assert True is policy.should_retry(500, False, 4)
    assert True is policy.should_retry(500, False, 5)
    assert False is policy.should_retry(500, False, 6)

    # Check based on status code
    assert False is policy.should_retry(201, False, 0)
    assert False is policy.should_retry(201, False, 6)

    # Check based on error
    assert True is policy.should_retry(201, True, 0)
    assert True is policy.should_retry(201, True, 6)


def test_with_default_values():
    times = conf.retry_seconds_to_sleep_list()
    max_retries = conf.configurable_retry_policy_max_retries()
    policy = ConfigurableRetryPolicy(times, max_retries)

    assert times[0] == policy.seconds_to_sleep(0)
    assert times[0] == policy.seconds_to_sleep(1)
    assert times[1] == policy.seconds_to_sleep(2)
    assert times[2] == policy.seconds_to_sleep(3)
    assert times[3] == policy.seconds_to_sleep(4)
    assert times[4] == policy.seconds_to_sleep(5)
    assert times[4] == policy.seconds_to_sleep(6)
    assert times[4] == policy.seconds_to_sleep(7)
    assert times[4] == policy.seconds_to_sleep(8)
    assert times[4] == policy.seconds_to_sleep(9)

    # Check based on retry count
    assert True is policy.should_retry(500, False, 0)
    assert True is policy.should_retry(500, False, 7)
    assert True is policy.should_retry(500, False, 8)
    assert False is policy.should_retry(500, False, 9)

    # Check based on status code
    assert False is policy.should_retry(201, False, 0)
    assert False is policy.should_retry(201, False, 9)

    # Check based on error
    assert True is policy.should_retry(201, True, 0)
    assert True is policy.should_retry(201, True, 9)


def test_with_negative_values():
    times = [0.1, -1]
    max_retries = 5

    try:
        policy = ConfigurableRetryPolicy(times, max_retries)
        assert False
    except BadUserConfigurationException:
        assert True



import pytest
from mock import MagicMock

import sparkmagic.utils.configuration as conf
from sparkmagic.livyclientlib.exceptions import BadUserConfigurationException
from sparkmagic.utils.constants import AUTH_BASIC, NO_AUTH


def setup_function():
    conf.override_all({})


def test_configuration_override_base64_password():
    kpc = {
        "username": "U",
        "password": "P",
        "base64_password": "cGFzc3dvcmQ=",
        "url": "L",
        "auth": AUTH_BASIC,
    }
    overrides = {conf.kernel_python_credentials.__name__: kpc}
    conf.override_all(overrides)
    conf.override(conf.livy_session_startup_timeout_seconds.__name__, 1)
    assert conf.d == {
        conf.kernel_python_credentials.__name__: kpc,
        conf.livy_session_startup_timeout_seconds.__name__: 1,
    }
    assert conf.livy_session_startup_timeout_seconds() == 1
    assert conf.base64_kernel_python_credentials() == {
        "username": "U",
        "password": "password",
        "url": "L",
        "auth": AUTH_BASIC,
    }


def test_configuration_auth_missing_basic_auth():
    kpc = {"username": "U", "password": "P", "url": "L"}
    overrides = {conf.kernel_python_credentials.__name__: kpc}
    conf.override_all(overrides)
    assert conf.base64_kernel_python_credentials() == {
        "username": "U",
        "password": "P",
        "url": "L",
        "auth": AUTH_BASIC,
    }


def test_configuration_auth_missing_no_auth():
    kpc = {"username": "", "password": "", "url": "L"}
    overrides = {conf.kernel_python_credentials.__name__: kpc}
    conf.override_all(overrides)
    assert conf.base64_kernel_python_credentials() == {
        "username": "",
        "password": "",
        "url": "L",
        "auth": NO_AUTH,
    }


def test_configuration_override_fallback_to_password():
    kpc = {"username": "U", "password": "P", "url": "L", "auth": NO_AUTH}
    overrides = {conf.kernel_python_credentials.__name__: kpc}
    conf.override_all(overrides)
    conf.override(conf.livy_session_startup_timeout_seconds.__name__, 1)
    assert conf.d == {
        conf.kernel_python_credentials.__name__: kpc,
        conf.livy_session_startup_timeout_seconds.__name__: 1,
    }
    assert conf.livy_session_startup_timeout_seconds() == 1
    assert conf.base64_kernel_python_credentials() == kpc


def test_configuration_override_work_with_empty_password():
    kpc = {
        "username": "U",
        "base64_password": "",
        "password": "",
        "url": "",
        "auth": AUTH_BASIC,
    }
    overrides = {conf.kernel_python_credentials.__name__: kpc}
    conf.override_all(overrides)
    conf.override(conf.livy_session_startup_timeout_seconds.__name__, 1)
    assert conf.d == {
        conf.kernel_python_credentials.__name__: kpc,
        conf.livy_session_startup_timeout_seconds.__name__: 1,
    }
    assert conf.livy_session_startup_timeout_seconds() == 1
    assert conf.base64_kernel_python_credentials() == {
        "username": "U",
        "password": "",
        "url": "",
        "auth": AUTH_BASIC,
    }


def test_configuration_raise_error_for_bad_base64_password():
    with pytest.raises(BadUserConfigurationException):
        kpc = {"username": "U", "base64_password": "P", "url": "L"}
        overrides = {conf.kernel_python_credentials.__name__: kpc}
        conf.override_all(overrides)
        conf.override(conf.livy_session_startup_timeout_seconds.__name__, 1)
        conf.base64_kernel_python_credentials()


def test_share_config_between_pyspark_and_pyspark3():
    kpc = {
        "username": "U",
        "password": "P",
        "base64_password": "cGFzc3dvcmQ=",
        "url": "L",
        "auth": AUTH_BASIC,
    }
    overrides = {conf.kernel_python_credentials.__name__: kpc}
    assert (
        conf.base64_kernel_python3_credentials()
        == conf.base64_kernel_python_credentials()
    )


def test_get_session_properties():
    assert conf.get_session_properties("python") == {"kind": "pyspark"}


def test_get_session_properties_no_defaults():
    conf.override(
        conf.session_configs.__name__, {"foo": "bar", "config": {"foo": "bar"}}
    )
    assert conf.get_session_properties("python") == {
        "kind": "pyspark",
        "foo": "bar",
        "config": {"foo": "bar"},
    }


def test_get_session_properties_with_defaults():
    conf.override(
        conf.session_configs_defaults.__name__,
        {"foo": "default", "config": {"foo": "default", "default": "default"}},
    )
    assert conf.get_session_properties("python") == {
        "kind": "pyspark",
        "foo": "default",
        "config": {"foo": "default", "default": "default"},
    }


def test_get_session_properties_with_defaults_and_overides():
    conf.override(
        conf.session_configs.__name__,
        {
            "foobar": "foobar",
            "foo": "bar",
            "config": {"foo": "bar"},
            "l1": {
                "l1k1": {
                    "l2k1": "bar",
                    "l2k3": "bar",
                },
                "l1k2": "bar",
                "l1k4": "bar",
            },
        },
    )
    conf.override(
        conf.session_configs_defaults.__name__,
        {
            "foo": "default",
            "l1": {
                "l1k1": {"l2k1": "default", "l2k2": "default"},
                "l1k2": "default",
                "l1k3": "default",
            },
            "config": {"foo": "bar", "default": "default"},
        },
    )
    assert conf.get_session_properties("python") == {
        "kind": "pyspark",
        "foobar": "foobar",
        "foo": "bar",
        "config": {"foo": "bar", "default": "default"},
        "l1": {
            "l1k1": {"l2k1": "bar", "l2k2": "default", "l2k3": "bar"},
            "l1k2": "bar",
            "l1k3": "default",
            "l1k4": "bar",
        },
    }



import json
from mock import MagicMock, call
import pytest

import sparkmagic.utils.constants as constants
import sparkmagic.utils.configuration as conf
from sparkmagic.livyclientlib.exceptions import (
    LivyClientTimeoutException,
    LivyUnexpectedStatusException,
    BadUserDataException,
    SqlContextNotFoundException,
)
from sparkmagic.livyclientlib.livysession import LivySession


class DummyResponse:
    def __init__(self, status_code, json_text):
        self._status_code = status_code
        self._json_text = json_text

    def json(self):
        return json.loads(self._json_text)

    @property
    def status_code(self):
        return self._status_code


class TestLivySession(object):
    _multiprocess_can_split_ = True

    pi_result = "Pi is roughly 3.14336"

    session_create_json = json.loads(
        '{"id":0,"state":"starting","kind":"spark","log":[]}'
    )
    resource_limit_json = json.loads(
        '{"id":0,"state":"starting","kind":"spark","log":['
        '"Queue\'s AM resource limit exceeded."]}'
    )
    ready_sessions_json = json.loads(
        '{"id":0,"state":"idle","kind":"spark","log":[""]}'
    )
    recovering_sessions_json = json.loads(
        '{"id":0,"state":"recovering","kind":"spark","log":[""]}'
    )
    error_sessions_json = json.loads(
        '{"id":0,"state":"error","kind":"spark","log":[""]}'
    )
    busy_sessions_json = json.loads('{"id":0,"state":"busy","kind":"spark","log":[""]}')
    post_statement_json = json.loads('{"id":0,"state":"running","output":null}')
    waiting_statement_json = json.loads('{"id":0,"state":"waiting","output":null}')
    running_statement_json = json.loads('{"id":0,"state":"running","output":null}')
    ready_statement_json = json.loads(
        '{"id":0,"state":"available","output":{"status":"ok",'
        '"execution_count":0,"data":{"text/plain":"Pi is roughly 3.14336"}}}'
    )
    ready_statement_null_output_json = json.loads(
        '{"id":0,"state":"available","output":null}'
    )
    ready_statement_failed_json = json.loads(
        '{"id":0,"state":"available","output":{"status":"error",'
        '"evalue":"error","traceback":"error"}}'
    )
    log_json = json.loads('{"id":6,"from":0,"total":212,"log":["hi","hi"]}')

    def __init__(self):
        self.http_client = None
        self.spark_events = None

        self.get_statement_responses = []
        self.post_statement_responses = []
        self.get_session_responses = []
        self.post_session_responses = []

    def setup(self):
        self.http_client = MagicMock()
        self.spark_events = MagicMock()
        self.heartbeat_thread = MagicMock()

    def _next_statement_response_get(self, *args):
        val = self.get_statement_responses[0]
        self.get_statement_responses = self.get_statement_responses[1:]
        return val

    def _next_statement_response_post(self, *args):
        val = self.post_statement_responses[0]
        self.post_statement_responses = self.post_statement_responses[1:]
        return val

    def _next_session_response_get(self, *args):
        val = self.get_session_responses[0]
        self.get_session_responses = self.get_session_responses[1:]
        return val

    def _next_session_response_post(self, *args):
        val = self.post_session_responses[0]
        self.post_session_responses = self.post_session_responses[1:]
        return val

    def _create_session(
        self, kind=constants.SESSION_KIND_SPARK, session_id=-1, heartbeat_timeout=60
    ):
        ipython_display = MagicMock()
        session = LivySession(
            self.http_client,
            {"kind": kind},
            ipython_display,
            session_id,
            self.spark_events,
            heartbeat_timeout,
            self.heartbeat_thread,
        )
        return session

    def _create_session_with_fixed_get_response(self, get_session_json):
        self.http_client.get_session.return_value = get_session_json
        self.http_client.get_statement.return_value = self.ready_statement_json
        session = self._create_session()
        session.start()
        return session

    def test_doesnt_do_anything_or_create_sql_context_automatically(self):
        # If the session object does anything (attempts to create a session or run
        # a statement), the http_client will fail
        self.http_client = MagicMock(side_effect=ValueError)
        self._create_session()

    def test_constructor_starts_with_existing_session(self):
        session_id = 1
        session = self._create_session(session_id=session_id, heartbeat_timeout=0)

        assert session.id == session_id
        assert session._heartbeat_thread is None
        assert constants.LIVY_HEARTBEAT_TIMEOUT_PARAM not in list(
            session.properties.keys()
        )

    def test_constructor_starts_heartbeat_with_existing_session(self):
        conf.override_all({"heartbeat_refresh_seconds": 0.1})
        session_id = 1
        session = self._create_session(session_id=session_id)
        conf.override_all({})

        assert session.id == session_id
        assert self.heartbeat_thread.daemon
        self.heartbeat_thread.start.assert_called_once_with()
        assert not session._heartbeat_thread is None
        assert session.properties[constants.LIVY_HEARTBEAT_TIMEOUT_PARAM] > 0

    def test_start_with_heartbeat(self):
        self.http_client.post_session.return_value = self.session_create_json
        self.http_client.get_session.return_value = self.ready_sessions_json
        self.http_client.get_statement.return_value = self.ready_statement_json

        session = self._create_session()
        session.start()

        assert self.heartbeat_thread.daemon
        self.heartbeat_thread.start.assert_called_once_with()
        assert not session._heartbeat_thread is None
        assert session.properties[constants.LIVY_HEARTBEAT_TIMEOUT_PARAM] > 0

    def test_start_with_heartbeat_calls_only_once(self):
        self.http_client.post_session.return_value = self.session_create_json
        self.http_client.get_session.return_value = self.ready_sessions_json
        self.http_client.get_statement.return_value = self.ready_statement_json

        session = self._create_session()
        session.start()
        session.start()
        session.start()

        assert self.heartbeat_thread.daemon
        self.heartbeat_thread.start.assert_called_once_with()
        assert not session._heartbeat_thread is None

    def test_delete_with_heartbeat(self):
        self.http_client.post_session.return_value = self.session_create_json
        self.http_client.get_session.return_value = self.ready_sessions_json
        self.http_client.get_statement.return_value = self.ready_statement_json

        session = self._create_session()
        session.start()
        heartbeat_thread = session._heartbeat_thread

        session.delete()

        self.heartbeat_thread.stop.assert_called_once_with()
        assert session._heartbeat_thread is None

    def test_constructor_starts_with_no_session(self):
        session = self._create_session()

        assert session.id == -1

    def test_is_final_status(self):
        session = self._create_session()

        assert not session.is_final_status("idle")
        assert not session.is_final_status("starting")
        assert not session.is_final_status("busy")

        assert session.is_final_status("dead")
        assert session.is_final_status("error")
        assert session.is_final_status("success")
        assert session.is_final_status("killed")

    def test_start_scala_starts_session(self):
        self.http_client.post_session.return_value = self.session_create_json
        self.http_client.get_session.return_value = self.ready_sessions_json
        self.http_client.get_statement.return_value = self.ready_statement_json

        kind = constants.SESSION_KIND_SPARK
        session = self._create_session(kind=kind)
        session.start()

        assert kind == session.kind
        assert "idle" == session.status
        assert 0 == session.id
        self.http_client.post_session.assert_called_with(
            {"kind": "spark", "heartbeatTimeoutInSecond": 60}
        )

    def test_start_r_starts_session(self):
        self.http_client.post_session.return_value = self.session_create_json
        self.http_client.get_session.return_value = self.ready_sessions_json
        self.http_client.get_statement.return_value = self.ready_statement_json

        kind = constants.SESSION_KIND_SPARKR
        session = self._create_session(kind=kind)
        session.start()

        assert kind == session.kind
        assert "idle" == session.status
        assert 0 == session.id
        self.http_client.post_session.assert_called_with(
            {"kind": "sparkr", "heartbeatTimeoutInSecond": 60}
        )

    def test_start_python_starts_session(self):
        self.http_client.post_session.return_value = self.session_create_json
        self.http_client.get_session.return_value = self.ready_sessions_json
        self.http_client.get_statement.return_value = self.ready_statement_json

        kind = constants.SESSION_KIND_PYSPARK
        session = self._create_session(kind=kind)
        session.start()

        assert kind == session.kind
        assert "idle" == session.status
        assert 0 == session.id
        self.http_client.post_session.assert_called_with(
            {"kind": "pyspark", "heartbeatTimeoutInSecond": 60}
        )

    def test_start_passes_in_all_properties(self):
        self.http_client.post_session.return_value = self.session_create_json
        self.http_client.get_session.return_value = self.ready_sessions_json
        self.http_client.get_statement.return_value = self.ready_statement_json

        kind = constants.SESSION_KIND_SPARK
        properties = {"kind": kind, "extra": 1}

        ipython_display = MagicMock()
        session = LivySession(self.http_client, properties, ipython_display)
        session.start()

        self.http_client.post_session.assert_called_with(properties)

    def test_status_gets_latest_status(self):
        self.http_client.post_session.return_value = self.session_create_json
        self.http_client.get_session.return_value = self.ready_sessions_json
        self.http_client.get_statement.return_value = self.ready_statement_json
        session = self._create_session()
        session.start()

        session.refresh_status_and_info()
        state = session.status

        assert "idle" == state
        self.http_client.get_session.assert_called_with(0)

    def test_status_recovering(self):
        """Ensure 'recovering' state is supported: we go from recovering to
        idle."""
        self.http_client.post_session.return_value = self.session_create_json

        def get_session(i, calls=[]):
            if not calls:
                calls.append(1)
                return self.recovering_sessions_json
            else:
                return self.ready_sessions_json

        self.http_client.get_session.side_effect = get_session
        self.http_client.get_statement.return_value = self.ready_statement_json
        session = self._create_session()
        session.start()
        assert "idle" == session.status

    def test_logs_gets_latest_logs(self):
        self.http_client.post_session.return_value = self.session_create_json
        self.http_client.get_session.return_value = self.ready_sessions_json
        self.http_client.get_all_session_logs.return_value = self.log_json
        self.http_client.get_statement.return_value = self.ready_statement_json
        session = self._create_session()
        session.start()

        logs = session.get_logs()

        assert "hi\nhi" == logs
        self.http_client.get_all_session_logs.assert_called_with(0)

    def test_wait_for_idle_returns_when_in_state(self):
        self.http_client.post_session.return_value = self.session_create_json
        self.get_session_responses = [
            self.ready_sessions_json,
            self.ready_sessions_json,
            self.busy_sessions_json,
            self.ready_sessions_json,
        ]
        self.http_client.get_session.side_effect = self._next_session_response_get
        self.http_client.get_statement.return_value = self.ready_statement_json

        session = self._create_session()
        session.get_row_html = MagicMock()
        session.get_row_html.return_value = """<tr><td>row1</td></tr>"""

        session.start()

        session.wait_for_idle(30)

        self.http_client.get_session.assert_called_with(0)
        assert 4 == self.http_client.get_session.call_count

    def test_wait_for_idle_prints_resource_limit_message(self):
        self.http_client.post_session.return_value = self.session_create_json
        self.get_session_responses = [
            self.resource_limit_json,
            self.ready_sessions_json,
            self.ready_sessions_json,
            self.ready_sessions_json,
        ]
        self.http_client.get_session.side_effect = self._next_session_response_get
        self.http_client.get_statement.return_value = self.ready_statement_json
        self.http_client.get_all_session_logs.return_value = self.log_json

        session = self._create_session()
        session.get_row_html = MagicMock()
        session.get_row_html.return_value = """<tr><td>row1</td></tr>"""

        session.start()

        session.wait_for_idle(30)
        assert session.ipython_display.send_error.call_count == 1

    def test_wait_for_idle_throws_when_in_final_status(self):
        with pytest.raises(LivyUnexpectedStatusException):
            self.http_client.post_session.return_value = self.session_create_json
            self.get_session_responses = [
                self.ready_sessions_json,
                self.busy_sessions_json,
                self.busy_sessions_json,
                self.error_sessions_json,
            ]
            self.http_client.get_session.side_effect = self._next_session_response_get
            self.http_client.get_all_session_logs.return_value = self.log_json

            session = self._create_session()
            session.get_row_html = MagicMock()
            session.get_row_html.return_value = """<tr><td>row1</td></tr>"""

            session.start()

            session.wait_for_idle(30)

    def test_wait_for_idle_times_out(self):
        with pytest.raises(LivyClientTimeoutException):
            self.http_client.post_session.return_value = self.session_create_json
            self.get_session_responses = [
                self.ready_sessions_json,
                self.ready_sessions_json,
                self.busy_sessions_json,
                self.busy_sessions_json,
                self.ready_sessions_json,
            ]
            self.http_client.get_session.side_effect = self._next_session_response_get
            self.http_client.get_statement.return_value = self.ready_statement_json

            session = self._create_session()
            session.get_row_html = MagicMock()
            session.get_row_html.return_value = """<tr><td>row1</td></tr>"""

            session.start()

            session.wait_for_idle(0.01)

    def test_delete_session_when_active(self):
        self.http_client.post_session.return_value = self.session_create_json
        self.http_client.get_session.return_value = self.ready_sessions_json
        self.http_client.get_statement.return_value = self.ready_statement_json
        session = self._create_session()
        session.start()

        session.delete()

        assert "dead" == session.status

    def test_delete_session_when_not_started(self):
        self.http_client.post_session.return_value = self.session_create_json
        session = self._create_session()

        session.delete()

        assert session.ipython_display.send_error.call_count == 1

    def test_delete_session_when_dead_throws(self):
        self.http_client.post.return_value = self.session_create_json
        session = self._create_session()
        session.status = "dead"

        session.delete()

        assert session.ipython_display.send_error.call_count == 0

    def test_start_emits_start_end_session(self):
        self.http_client.post_session.return_value = self.session_create_json
        self.http_client.get_session.return_value = self.ready_sessions_json
        self.http_client.get_statement.return_value = self.ready_statement_json

        kind = constants.SESSION_KIND_SPARK
        session = self._create_session(kind=kind)
        session.start()

        self.spark_events.emit_session_creation_start_event.assert_called_once_with(
            session.guid, kind
        )
        self.spark_events.emit_session_creation_end_event.assert_called_once_with(
            session.guid, kind, session.id, session.status, True, "", ""
        )

    def test_start_emits_start_end_failed_session_when_bad_status(self):
        self.http_client.post_session.side_effect = ValueError
        self.http_client.get_session.return_value = self.ready_sessions_json

        kind = constants.SESSION_KIND_SPARK
        session = self._create_session(kind=kind)

        try:
            session.start()
            assert False
        except ValueError:
            pass

        self.spark_events.emit_session_creation_start_event.assert_called_once_with(
            session.guid, kind
        )
        self.spark_events.emit_session_creation_end_event.assert_called_once_with(
            session.guid, kind, session.id, session.status, False, "ValueError", ""
        )

    def test_start_emits_start_end_failed_session_when_wait_for_idle_throws(self):
        self.http_client.post_session.return_value = self.session_create_json
        self.http_client.get_session.return_value = self.ready_sessions_json

        kind = constants.SESSION_KIND_SPARK
        session = self._create_session(kind=kind)
        session.wait_for_idle = MagicMock(side_effect=ValueError)

        try:
            session.start()
            assert False
        except ValueError:
            pass

        self.spark_events.emit_session_creation_start_event.assert_called_once_with(
            session.guid, kind
        )
        self.spark_events.emit_session_creation_end_event.assert_called_once_with(
            session.guid, kind, session.id, session.status, False, "ValueError", ""
        )

    def test_delete_session_emits_start_end(self):
        self.http_client.post_session.return_value = self.session_create_json
        self.http_client.get_session.return_value = self.ready_sessions_json
        self.http_client.get_statement.return_value = self.ready_statement_json

        session = self._create_session()
        session.start()
        end_id = session.id
        end_status = constants.BUSY_SESSION_STATUS
        session.status = end_status

        session.delete()

        assert session.id == -1
        self.spark_events.emit_session_deletion_start_event.assert_called_once_with(
            session.guid, session.kind, end_id, end_status
        )
        self.spark_events.emit_session_deletion_end_event.assert_called_once_with(
            session.guid,
            session.kind,
            end_id,
            constants.DEAD_SESSION_STATUS,
            True,
            "",
            "",
        )

    def test_delete_session_emits_start_failed_end_when_delete_throws(self):
        self.http_client.delete_session.side_effect = ValueError
        self.http_client.get_session.return_value = self.ready_sessions_json
        self.http_client.get_statement.return_value = self.ready_statement_json

        session = self._create_session()
        session.start()
        session.id = 0
        end_id = session.id
        end_status = constants.BUSY_SESSION_STATUS
        session.status = end_status

        try:
            session.delete()
            assert False
        except ValueError:
            pass

        self.spark_events.emit_session_deletion_start_event.assert_called_once_with(
            session.guid, session.kind, end_id, end_status
        )
        self.spark_events.emit_session_deletion_end_event.assert_called_once_with(
            session.guid, session.kind, end_id, end_status, False, "ValueError", ""
        )

    def test_delete_session_emits_start_failed_end_when_in_bad_state(self):
        self.http_client.get_session.return_value = self.ready_sessions_json
        self.http_client.get_statement.return_value = self.ready_statement_json

        session = self._create_session()
        session.start()
        session.id = 0
        end_id = session.id
        end_status = constants.DEAD_SESSION_STATUS
        session.status = end_status

        session.delete()

        assert 0 == session.ipython_display.send_error.call_count
        self.spark_events.emit_session_deletion_start_event.assert_called_once_with(
            session.guid, session.kind, end_id, end_status
        )
        self.spark_events.emit_session_deletion_end_event.assert_called_once_with(
            session.guid,
            session.kind,
            end_id,
            constants.DEAD_SESSION_STATUS,
            True,
            "",
            "",
        )

    def test_get_empty_app_id(self):
        self._verify_get_app_id("null", None, 7)

    def test_get_missing_app_id(self):
        self._verify_get_app_id(None, None, 7)

    def test_get_normal_app_id(self):
        self._verify_get_app_id('"app_id_123"', "app_id_123", 6)

    def test_get_empty_driver_log_url(self):
        self._verify_get_driver_log_url("null", None)

    def test_get_normal_driver_log_url(self):
        self._verify_get_driver_log_url('"http://example.com"', "http://example.com")

    def test_missing_app_info_get_driver_log_url(self):
        self._verify_get_driver_log_url_json(self.ready_sessions_json, None)

    def _verify_get_app_id(self, mock_app_id, expected_app_id, expected_call_count):
        mock_field = ',"appId":' + mock_app_id if mock_app_id is not None else ""
        get_session_json = json.loads(
            '{"id":0,"state":"idle","output":null%s,"log":""}' % mock_field
        )
        session = self._create_session_with_fixed_get_response(get_session_json)

        app_id = session.get_app_id()

        assert expected_app_id == app_id
        assert expected_call_count == self.http_client.get_session.call_count

    def _verify_get_driver_log_url(self, mock_driver_log_url, expected_url):
        mock_field = (
            '"driverLogUrl":' + mock_driver_log_url
            if mock_driver_log_url is not None
            else ""
        )
        session_json = json.loads(
            '{"id":0,"state":"idle","output":null,"appInfo":{%s},"log":""}' % mock_field
        )
        self._verify_get_driver_log_url_json(session_json, expected_url)

    def _verify_get_driver_log_url_json(self, get_session_json, expected_url):
        session = self._create_session_with_fixed_get_response(get_session_json)

        driver_log_url = session.get_driver_log_url()

        assert expected_url == driver_log_url
        assert 7 == self.http_client.get_session.call_count

    def test_get_empty_spark_ui_url(self):
        self._verify_get_spark_ui_url("null", None)

    def test_get_normal_spark_ui_url(self):
        self._verify_get_spark_ui_url('"http://example.com"', "http://example.com")

    def test_missing_app_info_get_spark_ui_url(self):
        self._verify_get_spark_ui_url_json(self.ready_sessions_json, None)

    def _verify_get_spark_ui_url(self, mock_spark_ui_url, expected_url):
        mock_field = (
            '"sparkUiUrl":' + mock_spark_ui_url if mock_spark_ui_url is not None else ""
        )
        session_json = json.loads(
            '{"id":0,"state":"idle","output":null,"appInfo":{%s},"log":""}' % mock_field
        )
        self._verify_get_spark_ui_url_json(session_json, expected_url)

    def _verify_get_spark_ui_url_json(self, get_session_json, expected_url):
        session = self._create_session_with_fixed_get_response(get_session_json)

        spark_ui_url = session.get_spark_ui_url()

        assert expected_url == spark_ui_url
        assert 7 == self.http_client.get_session.call_count

    def test_get_row_html(self):
        session_id1 = 1
        session1 = self._create_session(session_id=session_id1)
        session1.get_app_id = MagicMock()
        session1.get_spark_ui_url = MagicMock()
        session1.get_driver_log_url = MagicMock()
        session1.get_user = MagicMock()
        session1.get_app_id.return_value = "app1234"
        session1.status = constants.IDLE_SESSION_STATUS
        session1.get_spark_ui_url.return_value = "https://microsoft.com/sparkui"
        session1.get_driver_log_url.return_value = "https://microsoft.com/driverlog"
        session1.get_user.return_value = "userTest"
        html1 = session1.get_row_html(1)

        assert (
            html1
            == """<tr><td>1</td><td>app1234</td><td>spark</td><td>idle</td><td><a target="_blank" href="https://microsoft.com/sparkui">Link</a></td><td><a target="_blank" href="https://microsoft.com/driverlog">Link</a></td><td>userTest</td><td>\u2714</td></tr>"""
        )

        session_id2 = 3
        session2 = self._create_session(
            kind=constants.SESSION_KIND_PYSPARK, session_id=session_id2
        )
        session2.get_app_id = MagicMock()
        session2.get_spark_ui_url = MagicMock()
        session2.get_driver_log_url = MagicMock()
        session2.get_user = MagicMock()
        session2.get_app_id.return_value = "app5069"
        session2.status = constants.BUSY_SESSION_STATUS
        session2.get_spark_ui_url.return_value = None
        session2.get_driver_log_url.return_value = None
        session2.get_user.return_value = "userTest2"

        html2 = session2.get_row_html(1)

        assert (
            html2
            == """<tr><td>3</td><td>app5069</td><td>pyspark</td><td>busy</td><td></td><td></td><td>userTest2</td><td></td></tr>"""
        )

    def test_link(self):
        url = "https://microsoft.com"
        assert (
            LivySession.get_html_link("Link", url)
            == """<a target="_blank" href="https://microsoft.com">Link</a>"""
        )

        url = None
        assert LivySession.get_html_link("Link", url) == ""

    def test_spark_session_available(self):
        self.http_client.post_session.return_value = self.session_create_json
        self.http_client.get_session.return_value = self.ready_sessions_json
        self.http_client.get_statement.return_value = self.ready_statement_json
        session = self._create_session()
        session.start()
        assert session.sql_context_variable_name == "spark"

    def test_sql_context_available(self):
        self.http_client.post_session.return_value = self.session_create_json
        self.http_client.get_session.return_value = self.ready_sessions_json
        self.get_statement_responses = [
            self.ready_statement_failed_json,
            self.ready_statement_json,
        ]
        self.http_client.get_statement.side_effect = self._next_statement_response_get
        session = self._create_session()
        session.start()
        assert session.sql_context_variable_name == "sqlContext"

    def test_spark_session_and_sql_context_unavailable(self):
        with pytest.raises(SqlContextNotFoundException):
            self.http_client.post_session.return_value = self.session_create_json
            self.http_client.get_session.return_value = self.ready_sessions_json
            self.http_client.get_statement.return_value = (
                self.ready_statement_failed_json
            )
            session = self._create_session()
            session.start()

    def test_is_posted(self):
        self.http_client.post_session.return_value = self.session_create_json
        self.http_client.get_session.return_value = self.ready_sessions_json
        self.http_client.get_statement.return_value = self.ready_statement_json

        session = self._create_session()
        assert not session.is_posted()
        session.start()
        assert session.is_posted()



from IPython.core.error import UsageError
from mock import MagicMock
import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal

from sparkmagic.livyclientlib.exceptions import BadUserDataException
from sparkmagic.utils.utils import parse_argstring_or_throw, records_to_dataframe
from sparkmagic.utils.constants import SESSION_KIND_PYSPARK
from sparkmagic.utils.dataframe_parser import (
    DataframeHtmlParser,
    cell_contains_dataframe,
    CellComponentType,
    cell_components_iter,
    CellOutputHtmlParser,
)

import unittest


def test_parse_argstring_or_throw():
    parse_argstring = MagicMock(side_effect=UsageError("OOGABOOGABOOGA"))
    try:
        parse_argstring_or_throw(
            MagicMock(), MagicMock(), parse_argstring=parse_argstring
        )
        assert False
    except BadUserDataException as e:
        assert str(e) == str(parse_argstring.side_effect)

    parse_argstring = MagicMock(side_effect=ValueError("AN UNKNOWN ERROR HAPPENED"))
    try:
        parse_argstring_or_throw(
            MagicMock(), MagicMock(), parse_argstring=parse_argstring
        )
        assert False
    except ValueError as e:
        assert e is parse_argstring.side_effect


def test_records_to_dataframe_missing_value_first():
    result = """{"z":100, "y":50}
{"z":25, "nullv":1.0, "y":10}"""

    df = records_to_dataframe(result, SESSION_KIND_PYSPARK, True)
    expected = pd.DataFrame(
        [{"z": 100, "nullv": None, "y": 50}, {"z": 25, "nullv": 1, "y": 10}],
        columns=["z", "nullv", "y"],
    )
    assert_frame_equal(expected, df)


def test_records_to_dataframe_coercing():
    result = """{"z":"100", "y":"2016-01-01"}
{"z":"25", "y":"2016-01-01"}"""

    df = records_to_dataframe(result, SESSION_KIND_PYSPARK, True)
    expected = pd.DataFrame(
        [
            {"z": 100, "y": np.datetime64("2016-01-01")},
            {"z": 25, "y": np.datetime64("2016-01-01")},
        ],
        columns=["z", "y"],
    )
    assert_frame_equal(expected, df)


def test_records_to_dataframe_no_coercing():
    result = """{"z":"100", "y":"2016-01-01"}
{"z":"25", "y":"2016-01-01"}"""

    df = records_to_dataframe(result, SESSION_KIND_PYSPARK, False)
    expected = pd.DataFrame(
        [{"z": "100", "y": "2016-01-01"}, {"z": "25", "y": "2016-01-01"}],
        columns=["z", "y"],
    )
    assert_frame_equal(expected, df)


def test_records_to_dataframe_missing_value_later():
    result = """{"z":25, "nullv":1.0, "y":10}
{"z":100, "y":50}"""

    df = records_to_dataframe(result, SESSION_KIND_PYSPARK, True)
    expected = pd.DataFrame(
        [{"z": 25, "nullv": 1, "y": 10}, {"z": 100, "nullv": None, "y": 50}],
        columns=["z", "nullv", "y"],
    )
    assert_frame_equal(expected, df)


class TestDataframeParsing(unittest.TestCase):
    def test_dataframe_component(self):
        cell = """
                +---+------+
                | id|animal|
                +---+------+
                |  1|   bat|
                |  2| mouse|
                |  3| horse|
                +---+------+

            Only showing the last 20 rows 
        """
        dc = DataframeHtmlParser(cell)
        rows = dc.row_iter()
        self.assertDictEqual(next(rows), {"id": "1", "animal": "bat"})
        self.assertDictEqual(next(rows), {"id": "2", "animal": "mouse"})
        self.assertDictEqual(next(rows), {"id": "3", "animal": "horse"})
        with self.assertRaises(StopIteration):
            next(rows)

        cell = """
                    +---+------+
                    | id|animal|
                    +---+------+
                    +---+------+
                """
        dc = DataframeHtmlParser(cell)
        rows = dc.row_iter()
        with self.assertRaises(StopIteration):
            next(rows)

        cell = """
                +---+------+
                | id|animal|
                +---+------+
                |  1|   bat|
                |  2| mouse |
                |  3| horse|
                +---+------+

            Only showing the last 20 rows 
        """
        dc = DataframeHtmlParser(cell)
        rows = dc.row_iter()
        self.assertDictEqual(next(rows), {"id": "1", "animal": "bat"})
        with self.assertRaises(ValueError):
            next(rows)

    def test_dataframe_parsing(self):
        cell = """+---+------+
| id|animal|
+---+------+
|  1|   bat|
|  2| mouse|
|  3| horse|
+---+------+

            Only showing the last 20 rows 
        """
        self.assertTrue(cell_contains_dataframe(cell))

        cell = """
                +---+------+
                | id|animal|
                +---+------+
                |  1|   bat|
                |  2| mouse|
                |  3| horse|
                +---+------+

            Only showing the last 20 rows 
        """
        self.assertTrue(
            cell_contains_dataframe(cell), "Matches with leading whitespaces"
        )

        cell = """
                +---+------+
                | id|animal|
                +---+------+
                |  1|   bat|
                |  2| mouse|
                |  3| horse|
                +---+------+

                +---+------+
                | id|animal|
                +---+------+
                |  1|   cat|
                |  2| couse|
                |  3| corse|
                +---+------+

                """
        self.assertTrue(
            cell_contains_dataframe(cell), "Cell contains multiple dataframes"
        )

        cell = """
                    +---+------+
                    | id|animal|
                    +---+------+
                    +---+------+
                """
        self.assertTrue(cell_contains_dataframe(cell), "Empty DF")

        cell = """
                    +---+
                    | id|
                    +---+
                    +---+
                """
        self.assertTrue(cell_contains_dataframe(cell), "Single Column DF")

        cell = """
                +---+------+
                | id|animal|
                +---+------+
                |  1|   bat|
                |  2| mouse|
                |  3| horse|
                +---+----/-+

                Only showing the last 20 rows 
                """
        self.assertFalse(cell_contains_dataframe(cell), "Footer contains a /")

    def test_cell_components_iter(self):
        cell = """+---+------+
| id|animal|
+---+------+
|  1|   bat|
|  2| mouse|
|  3| horse|
+---+------+

            Only showing the last 20 rows 
        """
        df_spans = cell_components_iter(cell)

        self.assertEqual(next(df_spans), (CellComponentType.DF, 0, 90))
        self.assertEqual(next(df_spans), (CellComponentType.TEXT, 90, 143))
        self.assertEqual(cell[90:143].strip(), "Only showing the last 20 rows")
        with self.assertRaises(StopIteration):
            next(df_spans)

        cell = """

                Random stuff at the start

                +---+------+
                | id|animal|
                +---+------+
                |  1|   bat|
                |  2| mouse|
                |  3| horse|
                +---+------+

                Random stuff in the middle

                +---+------+
                | id|animal|
                +---+------+
                |  1|   cat|
                |  2| couse|
                |  3| corse|
                +---+------+

                Random stuff at the end
                """
        df_spans = cell_components_iter(cell)

        self.assertEqual(next(df_spans), (CellComponentType.TEXT, 0, 45))
        self.assertEqual(cell[0:45].strip(), "Random stuff at the start")
        self.assertEqual(next(df_spans), (CellComponentType.DF, 45, 247))

        self.assertEqual(next(df_spans), (CellComponentType.TEXT, 247, 293))
        self.assertEqual(cell[247:293].strip(), "Random stuff in the middle")

        self.assertEqual(next(df_spans), (CellComponentType.DF, 293, 495))

        self.assertEqual(next(df_spans), (CellComponentType.TEXT, 495, 553))
        self.assertEqual(cell[495:553].strip(), "Random stuff at the end")

        with self.assertRaises(StopIteration):
            next(df_spans)

        cell = """

                Random stuff at the start

                """
        df_spans = cell_components_iter(cell)

        self.assertEqual(next(df_spans), (CellComponentType.TEXT, 0, len(cell)))
        with self.assertRaises(StopIteration):
            next(df_spans)

        cell = ""
        df_spans = cell_components_iter(cell)
        with self.assertRaises(StopIteration):
            next(df_spans)

    def test_output_html_parser(self):
        cell = ""
        self.assertEqual(CellOutputHtmlParser.to_html(cell), "")

        cell = "Some random text"
        self.assertEqual(
            CellOutputHtmlParser.to_html(cell), "<pre>Some random text</pre>"
        )

        cell = """

                Random stuff at the start

                +---+------+
                | id|animal|
                +---+------+
                |  1|   bat|
                |  2| mouse|
                |  3| horse|
                +---+------+

                Random stuff in the middle
                +---+------+
                | id|animal|
                +---+------+
                |  1|   cat|
                |  2| couse|
                |  3| corse|
                +---+------+

                Random stuff at the end
                """
        self.maxDiff = 1200
        pre = '<pre style="word-break: unset; background-color: unset;">'
        self.assertEqual(
            CellOutputHtmlParser.to_html(cell),
            (
                "<pre>Random stuff at the start</pre><br />"
                f"<table><tr><th>{pre}id</pre></th><th>{pre}animal</pre></th></tr>"
                f"<tr><td>{pre}1</pre></td><td>{pre}bat</pre></td></tr>"
                f"<tr><td>{pre}2</pre></td><td>{pre}mouse</pre></td></tr>"
                f"<tr><td>{pre}3</pre></td><td>{pre}horse</pre></td></tr>"
                "</table><br /><pre>Random stuff in the middle</pre><br />"
                f"<table><tr><th>{pre}id</pre></th><th>{pre}animal</pre></th></tr>"
                f"<tr><td>{pre}1</pre></td><td>{pre}cat</pre></td></tr>"
                f"<tr><td>{pre}2</pre></td><td>{pre}couse</pre></td></tr>"
                f"<tr><td>{pre}3</pre></td><td>{pre}corse</pre></td></tr>"
                "</table><br /><pre>Random stuff at the end</pre>"
            ),
        )


if __name__ == "__main__":
    unittest.main()



from mock import MagicMock
import pytest

import sparkmagic.utils.configuration as conf
from sparkmagic.livyclientlib.exceptions import *


self = None
ipython_display = None
logger = None


def setup_function():
    global self, ipython_display, logger
    self = MagicMock()
    ipython_display = self.ipython_display
    logger = self.logger
    conf.override_all({})


def test_handle_expected_exceptions():
    mock_method = MagicMock()
    mock_method.__name__ = "MockMethod"
    decorated = handle_expected_exceptions(mock_method)
    assert decorated.__name__ == mock_method.__name__

    result = decorated(self, 1, 2, 3)
    assert result == mock_method.return_value
    assert ipython_display.send_error.call_count == 0
    mock_method.assert_called_once_with(self, 1, 2, 3)


def test_handle_expected_exceptions_handle():
    conf.override_all({"all_errors_are_fatal": False})
    mock_method = MagicMock(side_effect=LivyUnexpectedStatusException("ridiculous"))
    mock_method.__name__ = "MockMethod2"
    decorated = handle_expected_exceptions(mock_method)
    assert decorated.__name__ == mock_method.__name__

    result = decorated(self, 1, kwarg="foo")
    assert result is None
    assert ipython_display.send_error.call_count == 1
    mock_method.assert_called_once_with(self, 1, kwarg="foo")


def test_handle_expected_exceptions_throw():
    with pytest.raises(ValueError):
        mock_method = MagicMock(side_effect=ValueError("HALP"))
        mock_method.__name__ = "mock_meth"
        decorated = handle_expected_exceptions(mock_method)
        assert decorated.__name__ == mock_method.__name__

        _ = decorated(self, 1, kwarg="foo")


def test_handle_expected_exceptions_throws_if_all_errors_fatal():
    with pytest.raises(LivyUnexpectedStatusException):
        conf.override_all({"all_errors_are_fatal": True})
        mock_method = MagicMock(side_effect=LivyUnexpectedStatusException("Oh no!"))
        mock_method.__name__ = "mock_meth"
        decorated = handle_expected_exceptions(mock_method)
        assert decorated.__name__ == mock_method.__name__

        _ = decorated(self, 1, kwarg="foo")


# test wrap with unexpected to true
def test_wrap_unexpected_exceptions():
    mock_method = MagicMock()
    mock_method.__name__ = "tos"
    decorated = wrap_unexpected_exceptions(mock_method)
    assert decorated.__name__ == mock_method.__name__

    result = decorated(self, 0.0)
    assert result == mock_method.return_value
    assert ipython_display.send_error.call_count == 0
    mock_method.assert_called_once_with(self, 0.0)


def test_wrap_unexpected_exceptions_handle():
    mock_method = MagicMock(side_effect=ValueError("~~~~~~"))
    mock_method.__name__ = "tos"
    decorated = wrap_unexpected_exceptions(mock_method)
    assert decorated.__name__ == mock_method.__name__

    result = decorated(self, "FOOBAR", FOOBAR="FOOBAR")
    assert result is None
    assert ipython_display.send_error.call_count == 1
    mock_method.assert_called_once_with(self, "FOOBAR", FOOBAR="FOOBAR")


def test_wrap_unexpected_exceptions_throws_if_all_errors_fatal():
    with pytest.raises(ValueError):
        conf.override_all({"all_errors_are_fatal": True})
        mock_method = MagicMock(side_effect=ValueError("~~~~~~"))
        mock_method.__name__ = "tos"
        decorated = wrap_unexpected_exceptions(mock_method)
        assert decorated.__name__ == mock_method.__name__

        _ = decorated(self, "FOOBAR", FOOBAR="FOOBAR")



# Copyright (c) 2015  aggftw@gmail.com
# Distributed under the terms of the Modified BSD License.

from .linearretrypolicy import LinearRetryPolicy
from sparkmagic.livyclientlib.exceptions import BadUserConfigurationException


class ConfigurableRetryPolicy(LinearRetryPolicy):
    """Retry policy that returns a configurable number of seconds to sleep
    between calls, takes all status codes 500 or above to be retriable, and
    retries a given maximum number of times.

    If the retry count exceeds the number of items in the list, last
    item in the list is always returned.
    """

    def __init__(self, retry_seconds_to_sleep_list, max_retries):
        super(ConfigurableRetryPolicy, self).__init__(-1, max_retries)

        # If user configured to an empty list, let's make this behave as
        # a Linear Retry Policy by assigning a list of 1 element.
        if len(retry_seconds_to_sleep_list) == 0:
            retry_seconds_to_sleep_list = [5]
        elif not all(n > 0 for n in retry_seconds_to_sleep_list):
            raise BadUserConfigurationException(
                "All items in the list in your config need to be positive for configurable retry policy"
            )

        self.retry_seconds_to_sleep_list = retry_seconds_to_sleep_list
        self._max_index = len(self.retry_seconds_to_sleep_list) - 1

    def seconds_to_sleep(self, retry_count):
        index = max(retry_count - 1, 0)
        if index > self._max_index:
            index = self._max_index

        return self.retry_seconds_to_sleep_list[index]



"""Class for implementing a basic access authenticator for SparkMagic"""

from sparkmagic.livyclientlib.exceptions import BadUserDataException
from hdijupyterutils.ipywidgetfactory import IpyWidgetFactory
from requests.auth import HTTPBasicAuth
from .customauth import Authenticator


class Basic(HTTPBasicAuth, Authenticator):
    """Basic Access authenticator for SparkMagic"""

    def __init__(self, parsed_attributes=None):
        """Initializes the Authenticator with the attributes in the attributes
        parsed from a %spark magic command if applicable, or with default values
        otherwise.

        Args:
            self,
            parsed_attributes (IPython.core.magics.namespace): The namespace object that
            is created from parsing %spark magic command.
        """
        if parsed_attributes is not None:
            if parsed_attributes.user == "" or parsed_attributes.password == "":
                new_exc = BadUserDataException(
                    "Need to supply username and password arguments for "
                    "Basic Access Authentication. (e.g. -a username -p password)."
                )
                raise new_exc
            self.username = parsed_attributes.user
            self.password = parsed_attributes.password
        else:
            self.username = "username"
            self.password = "password"
        HTTPBasicAuth.__init__(self, self.username, self.password)
        Authenticator.__init__(self, parsed_attributes)

    def get_widgets(self, widget_width):
        """Creates and returns a list with an address, username, and password widget

        Args:
            widget_width (str): The width of all widgets to be created.

        Returns:
            Sequence[hdijupyterutils.ipywidgetfactory.IpyWidgetFactory]: list of widgets
        """
        ipywidget_factory = IpyWidgetFactory()

        self.user_widget = ipywidget_factory.get_text(
            description="Username:", value=self.username, width=widget_width
        )

        self.password_widget = ipywidget_factory.get_password(
            description="Password:", value=self.password, width=widget_width
        )

        widgets = [self.user_widget, self.password_widget]
        return Authenticator.get_widgets(self, widget_width) + widgets

    def update_with_widget_values(self):
        """Updates url, username, and password to be the value of their respective widgets."""
        Authenticator.update_with_widget_values(self)
        self.username = self.user_widget.value
        self.password = self.password_widget.value

    def __eq__(self, other):
        if not isinstance(other, Basic):
            return False
        return (
            self.url == other.url
            and self.username == other.username
            and self.password == other.password
        )

    def __call__(self, request):
        return HTTPBasicAuth.__call__(self, request)

    def __hash__(self):
        return hash((self.username, self.password, self.url, self.__class__.__name__))



from sparkmagic.livyclientlib.exceptions import BadUserDataException
from sparkmagic.livyclientlib.endpoint import Endpoint
from sparkmagic.auth.basic import Basic
from sparkmagic.auth.kerberos import Kerberos


def test_equality():
    basic_auth1 = Basic()
    basic_auth2 = Basic()
    kerberos_auth1 = Kerberos()
    kerberos_auth2 = Kerberos()
    assert Endpoint("http://url.com", basic_auth1) == Endpoint(
        "http://url.com", basic_auth2
    )
    assert Endpoint("http://url.com", kerberos_auth1) == Endpoint(
        "http://url.com", kerberos_auth2
    )


def test_inequality():
    basic_auth1 = Basic()
    basic_auth2 = Basic()
    basic_auth1.username = "user"
    basic_auth2.username = "different_user"
    assert Endpoint("http://url.com", basic_auth1) != Endpoint(
        "http://url.com", basic_auth2
    )


def test_invalid_url():
    basic_auth = Basic()
    try:
        endpoint = Endpoint(None, basic_auth)
        assert False
    except BadUserDataException:
        assert True



# Copyright (c) 2015  aggftw@gmail.com
# Distributed under the terms of the Modified BSD License.
from sparkmagic.controllerwidget.abstractmenuwidget import AbstractMenuWidget
from sparkmagic.controllerwidget.addendpointwidget import AddEndpointWidget
from sparkmagic.controllerwidget.manageendpointwidget import ManageEndpointWidget
from sparkmagic.controllerwidget.managesessionwidget import ManageSessionWidget
from sparkmagic.controllerwidget.createsessionwidget import CreateSessionWidget
from sparkmagic.livyclientlib.endpoint import Endpoint
from sparkmagic.utils.constants import LANGS_SUPPORTED
import sparkmagic.utils.configuration as conf
from sparkmagic.utils.utils import Namespace, initialize_auth


class MagicsControllerWidget(AbstractMenuWidget):
    def __init__(
        self, spark_controller, ipywidget_factory, ipython_display, endpoints=None
    ):
        super(MagicsControllerWidget, self).__init__(
            spark_controller, ipywidget_factory, ipython_display
        )

        if endpoints is None:
            endpoints = {
                endpoint.url: endpoint for endpoint in self._get_default_endpoints()
            }
        self.endpoints = endpoints

        self._refresh()

    def run(self):
        pass

    @staticmethod
    def _get_default_endpoints():
        default_endpoints = set()

        for kernel_type in LANGS_SUPPORTED:
            endpoint_config = getattr(conf, "kernel_%s_credentials" % kernel_type)()
            if (
                all([p in endpoint_config for p in ["url", "password", "username"]])
                and endpoint_config["url"] != ""
            ):
                user = endpoint_config["username"]
                passwd = endpoint_config["password"]
                args = Namespace(
                    user=user,
                    password=passwd,
                    auth=endpoint_config.get("auth", None),
                    url=endpoint_config.get("url", None),
                )
                auth_instance = initialize_auth(args)

                default_endpoints.add(
                    Endpoint(
                        auth=auth_instance,
                        url=endpoint_config["url"],
                        implicitly_added=True,
                    )
                )

        return default_endpoints

    def _refresh(self):
        dropdown_options = [(k, v) for k, v in self.endpoints.items()]
        self.endpoints_dropdown_widget = self.ipywidget_factory.get_dropdown(
            description="Endpoint:", options=dropdown_options
        )

        self.manage_session = ManageSessionWidget(
            self.spark_controller,
            self.ipywidget_factory,
            self.ipython_display,
            self._refresh,
        )
        self.create_session = CreateSessionWidget(
            self.spark_controller,
            self.ipywidget_factory,
            self.ipython_display,
            self.endpoints_dropdown_widget,
            self._refresh,
        )
        self.add_endpoint = AddEndpointWidget(
            self.spark_controller,
            self.ipywidget_factory,
            self.ipython_display,
            self.endpoints,
            self.endpoints_dropdown_widget,
            self._refresh,
        )
        self.manage_endpoint = ManageEndpointWidget(
            self.spark_controller,
            self.ipywidget_factory,
            self.ipython_display,
            self.endpoints,
            self._refresh,
        )

        self.tabs = self.ipywidget_factory.get_tab(
            children=[
                self.manage_session,
                self.create_session,
                self.add_endpoint,
                self.manage_endpoint,
            ]
        )
        self.tabs.set_title(0, "Manage Sessions")
        self.tabs.set_title(1, "Create Session")
        self.tabs.set_title(2, "Add Endpoint")
        self.tabs.set_title(3, "Manage Endpoints")

        self.children = [self.tabs]

        for child in self.children:
            child.parent_widget = self



from mock import MagicMock, patch
from tornado.concurrent import Future
from tornado.web import MissingArgumentError
from tornado.testing import gen_test
from tornado.testing import AsyncTestCase
import json

from sparkmagic.serverextension.handlers import ReconnectHandler
from sparkmagic.kernels.kernelmagics import KernelMagics
from sparkmagic.utils import constants


class SimpleObject(object):
    pass


class TestSparkMagicHandler(AsyncTestCase):
    reconnect_handler = None
    session_manager = None
    kernel_manager = None
    individual_kernel_manager = None
    client = None
    session_list = None
    spark_events = None
    path = "some_path.ipynb"
    kernel_id = "1"
    kernel_name = "pysparkkernel"
    session_id = "1"
    username = "username"
    password = "password"
    endpoint = "http://endpoint.com"
    auth = constants.AUTH_BASIC
    response_id = "0"
    good_msg = dict(content=dict(status="ok"))
    bad_msg = dict(content=dict(status="error", ename="SyntaxError", evalue="oh no!"))
    request = None

    def create_session_dict(self, path, kernel_id):
        return dict(
            notebook=dict(path=path),
            kernel=dict(id=kernel_id, name=self.kernel_name),
            id=self.session_id,
        )

    def get_argument(self, key):
        return dict(
            username=self.username,
            password=self.password,
            endpoint=self.endpoint,
            path=self.path,
        )[key]

    def setUp(self):
        # Mock kernel manager
        self.client = MagicMock()
        self.client.execute = MagicMock(return_value=self.response_id)
        self.client.get_shell_msg = MagicMock(return_value=self.good_msg)
        self.individual_kernel_manager = MagicMock()
        self.individual_kernel_manager.client = MagicMock(return_value=self.client)
        self.kernel_manager = MagicMock()
        self.kernel_manager.get_kernel = MagicMock(
            return_value=self.individual_kernel_manager
        )

        # Mock session manager
        self.session_list = [self.create_session_dict(self.path, self.kernel_id)]
        self.session_manager = MagicMock()
        self.session_manager.list_sessions = MagicMock(return_value=self.session_list)
        self.session_manager.create_session = MagicMock(
            return_value=self.create_session_dict(self.path, self.kernel_id)
        )

        # Mock spark events
        self.spark_events = MagicMock()

        # Mock request
        self.request = MagicMock()
        self.request.body = json.dumps(
            {
                "path": self.path,
                "username": self.username,
                "password": self.password,
                "endpoint": self.endpoint,
                "auth": self.auth,
            }
        )

        # Create mocked reconnect_handler
        ReconnectHandler.__bases__ = (SimpleObject,)
        self.reconnect_handler = ReconnectHandler()
        self.reconnect_handler.spark_events = self.spark_events
        self.reconnect_handler.session_manager = self.session_manager
        self.reconnect_handler.kernel_manager = self.kernel_manager
        self.reconnect_handler.set_status = MagicMock()
        self.reconnect_handler.finish = MagicMock()
        self.reconnect_handler.current_user = "alex"
        self.reconnect_handler.request = self.request
        self.reconnect_handler.logger = MagicMock()

        super(TestSparkMagicHandler, self).setUp()

    def test_msg_status(self):
        assert self.reconnect_handler._msg_status(self.good_msg) == "ok"
        assert self.reconnect_handler._msg_status(self.bad_msg) == "error"

    def test_msg_successful(self):
        assert self.reconnect_handler._msg_successful(self.good_msg) == True
        assert self.reconnect_handler._msg_successful(self.bad_msg) == False

    def test_msg_error(self):
        assert self.reconnect_handler._msg_error(self.good_msg) == None
        assert self.reconnect_handler._msg_error(self.bad_msg) == "{}:\n{}".format(
            "SyntaxError", "oh no!"
        )

    @gen_test
    def test_post_no_json(self):
        self.reconnect_handler.request.body = "{{}"

        res = yield self.reconnect_handler.post()
        assert res == None

        msg = "Invalid JSON in request body."
        self.reconnect_handler.set_status.assert_called_once_with(400)
        self.reconnect_handler.finish.assert_called_once_with(msg)
        self.spark_events.emit_cluster_change_event.assert_called_once_with(
            None, 400, False, msg
        )

    @gen_test
    def test_post_no_key(self):
        self.reconnect_handler.request.body = json.dumps({})

        res = yield self.reconnect_handler.post()
        assert res == None

        msg = "HTTP 400: Bad Request (Missing argument path)"
        self.reconnect_handler.set_status.assert_called_once_with(400)
        self.reconnect_handler.finish.assert_called_once_with(msg)
        self.spark_events.emit_cluster_change_event.assert_called_once_with(
            None, 400, False, msg
        )

    @patch("sparkmagic.serverextension.handlers.ReconnectHandler._get_kernel_manager")
    @gen_test
    def test_post_existing_kernel_with_auth_missing_no_auth(self, _get_kernel_manager):
        self.request.body = json.dumps(
            {
                "path": self.path,
                "username": "",
                "password": "",
                "endpoint": self.endpoint,
            }
        )
        kernel_manager_future = Future()
        kernel_manager_future.set_result(self.individual_kernel_manager)
        _get_kernel_manager.return_value = kernel_manager_future

        res = yield self.reconnect_handler.post()
        assert res == None

        code = "%{} -s {} -u {} -p {} -t {}".format(
            KernelMagics._do_not_call_change_endpoint.__name__,
            self.endpoint,
            "",
            "",
            constants.NO_AUTH,
        )
        self.client.execute.assert_called_once_with(
            code, silent=False, store_history=False
        )
        self.reconnect_handler.set_status.assert_called_once_with(200)
        self.reconnect_handler.finish.assert_called_once_with(
            '{"error": null, "success": true}'
        )
        self.spark_events.emit_cluster_change_event.assert_called_once_with(
            self.endpoint, 200, True, None
        )

    @patch("sparkmagic.serverextension.handlers.ReconnectHandler._get_kernel_manager")
    @gen_test
    def test_post_existing_kernel_with_auth_missing_basic_auth(
        self, _get_kernel_manager
    ):
        self.request.body = json.dumps(
            {
                "path": self.path,
                "username": self.username,
                "password": self.password,
                "endpoint": self.endpoint,
            }
        )
        kernel_manager_future = Future()
        kernel_manager_future.set_result(self.individual_kernel_manager)
        _get_kernel_manager.return_value = kernel_manager_future

        res = yield self.reconnect_handler.post()
        assert res == None

        code = "%{} -s {} -u {} -p {} -t {}".format(
            KernelMagics._do_not_call_change_endpoint.__name__,
            self.endpoint,
            self.username,
            self.password,
            constants.AUTH_BASIC,
        )
        self.client.execute.assert_called_once_with(
            code, silent=False, store_history=False
        )
        self.reconnect_handler.set_status.assert_called_once_with(200)
        self.reconnect_handler.finish.assert_called_once_with(
            '{"error": null, "success": true}'
        )
        self.spark_events.emit_cluster_change_event.assert_called_once_with(
            self.endpoint, 200, True, None
        )

    @patch("sparkmagic.serverextension.handlers.ReconnectHandler._get_kernel_manager")
    @gen_test
    def test_post_existing_kernel(self, _get_kernel_manager):
        kernel_manager_future = Future()
        kernel_manager_future.set_result(self.individual_kernel_manager)
        _get_kernel_manager.return_value = kernel_manager_future

        res = yield self.reconnect_handler.post()
        assert res == None

        code = "%{} -s {} -u {} -p {} -t {}".format(
            KernelMagics._do_not_call_change_endpoint.__name__,
            self.endpoint,
            self.username,
            self.password,
            self.auth,
        )
        self.client.execute.assert_called_once_with(
            code, silent=False, store_history=False
        )
        self.reconnect_handler.set_status.assert_called_once_with(200)
        self.reconnect_handler.finish.assert_called_once_with(
            '{"error": null, "success": true}'
        )
        self.spark_events.emit_cluster_change_event.assert_called_once_with(
            self.endpoint, 200, True, None
        )

    @patch("sparkmagic.serverextension.handlers.ReconnectHandler._get_kernel_manager")
    @gen_test
    def test_post_existing_kernel_failed(self, _get_kernel_manager):
        kernel_manager_future = Future()
        kernel_manager_future.set_result(self.individual_kernel_manager)
        _get_kernel_manager.return_value = kernel_manager_future
        self.client.get_shell_msg = MagicMock(return_value=self.bad_msg)

        res = yield self.reconnect_handler.post()
        assert res == None

        code = "%{} -s {} -u {} -p {} -t {}".format(
            KernelMagics._do_not_call_change_endpoint.__name__,
            self.endpoint,
            self.username,
            self.password,
            self.auth,
        )
        self.client.execute.assert_called_once_with(
            code, silent=False, store_history=False
        )
        self.reconnect_handler.set_status.assert_called_once_with(500)
        self.reconnect_handler.finish.assert_called_once_with(
            '{"error": "SyntaxError:\\noh no!", "success": false}'
        )
        self.spark_events.emit_cluster_change_event.assert_called_once_with(
            self.endpoint, 500, False, "SyntaxError:\noh no!"
        )

    @patch(
        "sparkmagic.serverextension.handlers.ReconnectHandler._get_kernel_manager_new_session"
    )
    @gen_test
    def test_get_kernel_manager_no_existing_kernel(
        self, _get_kernel_manager_new_session
    ):
        different_path = "different_path.ipynb"
        km_future = Future()
        km_future.set_result(self.individual_kernel_manager)
        _get_kernel_manager_new_session.return_value = km_future

        km = yield self.reconnect_handler._get_kernel_manager(
            different_path, self.kernel_name
        )

        assert self.individual_kernel_manager == km
        self.individual_kernel_manager.restart_kernel.assert_not_called()
        self.kernel_manager.get_kernel.assert_not_called()
        _get_kernel_manager_new_session.assert_called_once_with(
            different_path, self.kernel_name
        )

    @patch(
        "sparkmagic.serverextension.handlers.ReconnectHandler._get_kernel_manager_new_session"
    )
    @gen_test
    def test_get_kernel_manager_existing_kernel(self, _get_kernel_manager_new_session):
        km = yield self.reconnect_handler._get_kernel_manager(
            self.path, self.kernel_name
        )

        assert self.individual_kernel_manager == km
        self.individual_kernel_manager.restart_kernel.assert_called_once_with()
        _get_kernel_manager_new_session.assert_not_called()

    @patch(
        "sparkmagic.serverextension.handlers.ReconnectHandler._get_kernel_manager_new_session"
    )
    @gen_test
    def test_get_kernel_manager_different_kernel_type(
        self, _get_kernel_manager_new_session
    ):
        different_kernel = "sparkkernel"
        km_future = Future()
        km_future.set_result(self.individual_kernel_manager)
        _get_kernel_manager_new_session.return_value = km_future

        km = yield self.reconnect_handler._get_kernel_manager(
            self.path, different_kernel
        )

        assert self.individual_kernel_manager == km
        self.individual_kernel_manager.restart_kernel.assert_not_called()
        self.kernel_manager.get_kernel.assert_not_called()
        _get_kernel_manager_new_session.assert_called_once_with(
            self.path, different_kernel
        )
        self.session_manager.delete_session.assert_called_once_with(self.session_id)



from mock import MagicMock

from sparkmagic.livyclientlib.livyreliablehttpclient import LivyReliableHttpClient
from sparkmagic.livyclientlib.endpoint import Endpoint
import sparkmagic.utils.configuration as conf
import sparkmagic.utils.constants as constants
from sparkmagic.livyclientlib.exceptions import BadUserConfigurationException
from sparkmagic.livyclientlib.configurableretrypolicy import ConfigurableRetryPolicy
from sparkmagic.livyclientlib.linearretrypolicy import LinearRetryPolicy


def test_post_statement():
    http_client = MagicMock()
    livy_client = LivyReliableHttpClient(http_client, None)
    data = {"adlfj": "sadflkjsdf"}
    out = livy_client.post_statement(100, data)
    assert out == http_client.post.return_value.json.return_value
    http_client.post.assert_called_once_with("/sessions/100/statements", [201], data)


def test_get_statement():
    http_client = MagicMock()
    livy_client = LivyReliableHttpClient(http_client, None)
    out = livy_client.get_statement(100, 4)
    assert out == http_client.get.return_value.json.return_value
    http_client.get.assert_called_once_with("/sessions/100/statements/4", [200])


def test_cancel_statement():
    http_client = MagicMock()
    livy_client = LivyReliableHttpClient(http_client, None)
    out = livy_client.cancel_statement(100, 104)
    assert out == http_client.post.return_value.json.return_value
    http_client.post.assert_called_once_with(
        "/sessions/100/statements/104/cancel", [200], {}
    )


def test_get_sessions():
    http_client = MagicMock()
    livy_client = LivyReliableHttpClient(http_client, None)
    out = livy_client.get_sessions()
    assert out == http_client.get.return_value.json.return_value
    http_client.get.assert_called_once_with("/sessions", [200])


def test_post_session():
    http_client = MagicMock()
    livy_client = LivyReliableHttpClient(http_client, None)
    properties = {"adlfj": "sadflkjsdf", 1: [2, 3, 4, 5]}
    out = livy_client.post_session(properties)
    assert out == http_client.post.return_value.json.return_value
    http_client.post.assert_called_once_with("/sessions", [201], properties)


def test_get_session():
    http_client = MagicMock()
    livy_client = LivyReliableHttpClient(http_client, None)
    out = livy_client.get_session(4)
    assert out == http_client.get.return_value.json.return_value
    http_client.get.assert_called_once_with("/sessions/4", [200])


def test_delete_session():
    http_client = MagicMock()
    livy_client = LivyReliableHttpClient(http_client, None)
    livy_client.delete_session(99)
    http_client.delete.assert_called_once_with("/sessions/99", [200, 404])


def test_get_all_session_logs():
    http_client = MagicMock()
    livy_client = LivyReliableHttpClient(http_client, None)
    out = livy_client.get_all_session_logs(42)
    assert out == http_client.get.return_value.json.return_value
    http_client.get.assert_called_once_with("/sessions/42/log?from=0", [200])


def test_custom_headers():
    custom_headers = {"header1": "value1"}
    overrides = {conf.custom_headers.__name__: custom_headers}
    conf.override_all(overrides)
    endpoint = Endpoint("http://url.com", None)
    client = LivyReliableHttpClient.from_endpoint(endpoint)
    headers = client.get_headers()
    assert len(headers) == 2
    assert ("Content-Type" in headers) == True
    assert ("header1" in headers) == True


def test_retry_policy():
    # Default is configurable retry
    times = conf.retry_seconds_to_sleep_list()
    max_retries = conf.configurable_retry_policy_max_retries()
    policy = LivyReliableHttpClient._get_retry_policy()
    assert type(policy) is ConfigurableRetryPolicy
    assert times == policy.retry_seconds_to_sleep_list
    assert max_retries == policy.max_retries

    # Configure to linear retry
    _override_policy(constants.LINEAR_RETRY)
    policy = LivyReliableHttpClient._get_retry_policy()
    assert type(policy) is LinearRetryPolicy
    assert 5 == policy.seconds_to_sleep(1)
    assert 5 == policy.max_retries

    # Configure to something invalid
    _override_policy("garbage")
    try:
        policy = LivyReliableHttpClient._get_retry_policy()
        assert False
    except BadUserConfigurationException:
        assert True


def _override_policy(policy):
    overrides = {conf.retry_policy.__name__: policy}
    conf.override_all(overrides)



# coding=utf-8
from sparkmagic.kernels.wrapperkernel.usercodeparser import UserCodeParser
from sparkmagic.kernels.kernelmagics import KernelMagics


def test_empty_string():
    parser = UserCodeParser()

    assert "" == parser.get_code_to_run("")


def test_spark_code():
    parser = UserCodeParser()
    cell = "my code\nand more"

    assert "%%spark\nmy code\nand more" == parser.get_code_to_run(cell)


def test_local_single():
    parser = UserCodeParser()
    cell = """%local
hi
hi
hi"""

    assert "hi\nhi\nhi" == parser.get_code_to_run(cell)


def test_local_double():
    parser = UserCodeParser()
    cell = """%%local
hi
hi
hi"""

    assert "hi\nhi\nhi" == parser.get_code_to_run(cell)


def test_our_line_magics():
    parser = UserCodeParser()
    magic_name = KernelMagics.info.__name__
    cell = "%{}".format(magic_name)

    assert "%%{}\n ".format(magic_name) == parser.get_code_to_run(cell)


def test_our_line_magics_with_content():
    parser = UserCodeParser()
    magic_name = KernelMagics.info.__name__
    cell = """%{}
my content
more content""".format(
        magic_name
    )

    assert "%%{}\nmy content\nmore content\n ".format(
        magic_name
    ) == parser.get_code_to_run(cell)


def test_other_cell_magic():
    parser = UserCodeParser()
    cell = """%%magic
hi
hi
hi"""

    assert "{}".format(cell) == parser.get_code_to_run(cell)


def test_other_line_magic():
    parser = UserCodeParser()
    cell = """%magic
hi
hi
hi"""

    assert cell == parser.get_code_to_run(cell)


def test_scala_code():
    parser = UserCodeParser()
    cell = """/* Place the cursor in the cell and press SHIFT + ENTER to run */

val fruits = sc.textFile("wasb:///example/data/fruits.txt")
val yellowThings = sc.textFile("wasb:///example/data/yellowthings.txt")"""

    assert "%%spark\n{}".format(cell) == parser.get_code_to_run(cell)


def test_unicode():
    parser = UserCodeParser()
    cell = "print 'è🐙🐙🐙🐙'"

    assert "%%spark\n{}".format(cell) == parser.get_code_to_run(cell)


def test_unicode_in_magics():
    parser = UserCodeParser()
    magic_name = KernelMagics.info.__name__
    cell = """%{}
my content è🐙
more content""".format(
        magic_name
    )

    assert "%%{}\nmy content è🐙\nmore content\n ".format(
        magic_name
    ) == parser.get_code_to_run(cell)


def test_unicode_in_double_magics():
    parser = UserCodeParser()
    magic_name = KernelMagics.info.__name__
    cell = """%%{}
my content è🐙
more content""".format(
        magic_name
    )

    assert "%%{}\nmy content è🐙\nmore content\n ".format(
        magic_name
    ) == parser.get_code_to_run(cell)
